"""Add slack token field to organisation

Revision ID: 7bc746997e8d
Revises: 5c0498e264bc
Create Date: 2019-08-05 17:47:01.203644

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "7bc746997e8d"
down_revision = "5c0498e264bc"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "organisation",
        sa.Column("chat_token_slack", sa.String(length=80), nullable=True),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("organisation", "chat_token_slack")



"""Allow to set float values for Organisation.hours_by.day, Task.duration, Task.estimation and TimeSpent.duration

Revision ID: 8fbd40afbe5f
Revises: 2baede80b111
Create Date: 2023-02-23 14:47:29.417410

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "8fbd40afbe5f"
down_revision = "2baede80b111"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "organisation",
        "hours_by_day",
        existing_type=sa.INTEGER(),
        type_=sa.Float(),
        existing_nullable=False,
    )
    op.alter_column(
        "task",
        "duration",
        existing_type=sa.INTEGER(),
        type_=sa.Float(),
        existing_nullable=True,
    )
    op.alter_column(
        "task",
        "estimation",
        existing_type=sa.INTEGER(),
        type_=sa.Float(),
        existing_nullable=True,
    )
    op.alter_column(
        "time_spent",
        "duration",
        existing_type=sa.INTEGER(),
        type_=sa.Float(),
        existing_nullable=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "time_spent",
        "duration",
        existing_type=sa.Float(),
        type_=sa.INTEGER(),
        existing_nullable=False,
    )
    op.alter_column(
        "task",
        "estimation",
        existing_type=sa.Float(),
        type_=sa.INTEGER(),
        existing_nullable=True,
    )
    op.alter_column(
        "task",
        "duration",
        existing_type=sa.Float(),
        type_=sa.INTEGER(),
        existing_nullable=True,
    )
    op.alter_column(
        "organisation",
        "hours_by_day",
        existing_type=sa.Float(),
        type_=sa.INTEGER(),
        existing_nullable=False,
    )
    # ### end Alembic commands ###



"""add preview download flag to projects

Revision ID: f5bdca075cdc
Revises: f4ff5a73d283
Create Date: 2023-08-29 21:18:38.791600

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "f5bdca075cdc"
down_revision = "f4ff5a73d283"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("project", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "is_preview_download_allowed", sa.Boolean(), nullable=True
            )
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("project", schema=None) as batch_op:
        batch_op.drop_column("is_preview_download_allowed")

    # ### end Alembic commands ###



import math
import orjson as json
import sqlalchemy.orm as orm

from zou.app import app
from zou.app.utils import fields, string
from sqlalchemy import func
from sqlalchemy.inspection import inspect


def get_query_criterions_from_request(request):
    """
    Turn request parameters into a dict where keys are attributes to filter and
    values are values to filter.
    """
    criterions = {}
    for key, value in request.args.items():
        if key not in ["page"]:
            criterions[key] = value
    return criterions


def apply_criterions_to_db_query(model, db_query, criterions):
    """
    Apply criterions given in HTTP request to the sqlachemy db query object.
    """

    many_join_filter = []
    in_filter = []
    name_filter = []
    filters = {}

    column_names = inspect(model).all_orm_descriptors.keys()
    for key, value in criterions.items():
        if key not in ["page", "relations"] and key in column_names:
            field_key = getattr(model, key)

            is_many_to_many_field = hasattr(
                field_key, "property"
            ) and isinstance(
                field_key.property, orm.properties.RelationshipProperty
            )
            value_is_list = (
                hasattr(value, "__len__")
                and len(value) > 0
                and value[0] == "["
            )

            if key == "name" and field_key is not None:
                name_filter.append(value)

            elif is_many_to_many_field:
                many_join_filter.append((key, value))

            elif value_is_list:
                value_array = json.loads(value)
                in_filter.append(
                    field_key.in_(
                        [cast_value(value, field_key) for value in value_array]
                    )
                )
            else:
                filters[key] = cast_value(value, field_key)

        if filters:
            db_query = db_query.filter_by(**filters)

        for value in name_filter:
            db_query = db_query.filter(model.name.ilike(value))

        for id_filter in in_filter:
            db_query = db_query.filter(id_filter)

        for key, value in many_join_filter:
            db_query = db_query.filter(getattr(model, key).any(id=value))

    return db_query


def get_paginated_results(query, page, limit=None, relations=False):
    """
    Apply pagination to the query object.
    """
    if page < 1:
        entries = query.all()
        return fields.serialize_models(entries, relations=relations)
    else:
        limit = limit or app.config["NB_RECORDS_PER_PAGE"]
        total = query.count()
        offset = (page - 1) * limit

        nb_pages = int(math.ceil(total / float(limit)))
        query = query.limit(limit)
        query = query.offset(offset)

        if total < offset:
            result = {
                "data": [],
                "total": 0,
                "nb_pages": nb_pages,
                "limit": limit,
                "offset": offset,
                "page": page,
            }
        else:
            models = fields.serialize_models(query.all(), relations=relations)
            result = {
                "data": models,
                "total": total,
                "nb_pages": nb_pages,
                "limit": limit,
                "offset": offset,
                "page": page,
            }
        return result


def get_cursor_results(
    model,
    query,
    cursor_created_at,
    limit=None,
    relations=False,
):
    """ """
    limit = limit or app.config["NB_RECORDS_PER_PAGE"]
    total = query.count()
    query = (
        query.filter(model.created_at > cursor_created_at)
        .order_by(model.created_at, model.updated_at, model.id)
        .limit(limit)
    )
    models = fields.serialize_models(
        query.all(), relations=relations, milliseconds=True
    )
    result = {
        "data": models,
        "total": total,
        "limit": limit,
    }
    return result


def apply_sort_by(model, query, sort_by):
    """
    Apply an order by clause to a sqlalchemy query from a string parameter.
    """
    if sort_by in model.__table__.columns.keys():
        sort_field = model.__table__.columns[sort_by]
        if sort_by in ["created_at", "updated_at"]:
            sort_field = sort_field.desc()
    else:
        sort_field = model.updated_at.desc()
    return query.order_by(sort_field)


def cast_value(value, field_key):
    if field_key.type.python_type is bool:
        return string.strtobool(value)
    else:
        return func.cast(value, field_key.type)



from zou import __version__

swagger_config = {
    "headers": [
        ("Access-Control-Allow-Origin", "*"),
        ("Access-Control-Allow-Methods", "GET, POST, PUT, DELETE, OPTIONS"),
        ("Access-Control-Allow-Credentials", "true"),
        (
            "Access-Control-Allow-Headers",
            "Authorization, Origin, X-Requested-With, Content-Type, Accept",
        ),
    ],
    "specs": [{"endpoint": "openapi", "route": "/openapi.json"}],
    "static_url_path": "/docs",
    "swagger_ui": True,
    "specs_route": "/apidocs/",
    "openapi": "3.0.2",
}


description = f"""
## Welcome to the Kitsu API specification
```Version: {__version__}```

The Kitsu API allows to store and manage the data of your animation/VFX production. Through it you can link all the tools of your pipeline and make sure they are all synchronized.

An easy to use Python client to access this API is available:
[Python Kitsu Client documentation](https://gazu.cg-wire.com/)

## Authentication<div class="auth">

<p>Before you can use any of the endpoints outline below,
you will have to get a JWT token to authorize your requests.
</p>

<p>
You will find the information to retrieve it in the 
<a href="#tag/Authentication">Zou documentation</a>.
</p>

[OpenAPI definition](/openapi.json)
"""

swagger_template = {
    "openapi": "3.0.2",
    "info": {
        "title": "Kitsu API",
        "description": description,
        "contact": {
            "name": "CGWire",
            "url": "https://www.cg-wire.com",
        },
        "version": __version__,
        "license": {
            "name": "AGPL 3.0",
            "url": "https://www.gnu.org/licenses/agpl-3.0.en.html",
        },
    },
    "externalDocs": {
        "description": "Read the installation documentation",
        "url": "https://zou.cg-wire.com",
    },
    "host": "localhost:8080",
    "basePath": "/api",
    "schemes": ["http", "https"],
    "components": {
        "securitySchemes": {
            "JWT Authorization": {
                "name": "Authorization",
                "in": "header",
                "type": "apiKey",
                "description": "Format in header: **Authorization: Bearer {token}**. \n\n Value example: Bearer xxxxx.yyyyy.zzzzz",
            }
        }
    },
    "security": [{"JWT Authorization": []}],
    "tags": [
        {"name": "Authentication"},
        {"name": "Assets"},
        {"name": "Breakdown"},
        {"name": "Comments"},
        {"name": "Crud"},
        {"name": "Edits"},
        {"name": "Entities"},
        {"name": "Events"},
        {"name": "Export"},
        {"name": "Files"},
        {"name": "Import"},
        {"name": "Index"},
        {"name": "News"},
        {"name": "Persons"},
        {"name": "Playlists"},
        {"name": "Previews"},
        {"name": "Projects"},
        {"name": "Search"},
        {"name": "Shots"},
        {"name": "Tasks"},
        {"name": "User"},
    ],
    "definitions": {
        " Common fields for all model instances": {
            "type": "object",
            "properties": {
                " id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "A unique ID made of letters, hyphens and numbers",
                    "example": "a24a6ea4-ce75-4665-a070-57453082c25",
                },
                "created_at": {
                    "type": "string",
                    "format": "date-time",
                    "description": "The creation date",
                },
                "updated_at": {
                    "type": "string",
                    "format": "date-time",
                    "description": "The update date",
                },
            },
        },
        "Asset": {
            "type": "object",
            "properties": {
                "name": {"type": "string", "description": "Name of asset"},
                "code": {
                    "type": "string",
                    "description": "Utility field for the pipeline to identify the asset",
                },
                "description": {
                    "type": "string",
                    "description": "Asset brief",
                },
                "canceled": {
                    "type": "boolean",
                    "default": "False",
                    "description": "True if the asset has been delete one time, False otherwise",
                },
                "project_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Project ID",
                },
                "entity_type_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Asset type ID",
                },
                "source_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Field uset to set the episode_id",
                },
                "preview_file_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "ID of preview file used as thumbnail",
                },
                "data": {
                    "type": "string",
                    "format": "json",
                    "description": "Free JSON field to add metadata",
                },
                "shotgun_id": {
                    "type": "integer",
                    "description": "Used for synchronization with a Shotgun instance",
                },
            },
        },
        "AssetInstance": {
            "type": "object",
            "properties": {
                "asset_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Instantiated asset",
                },
                "name": {"type": "string"},
                "number": {"type": "integer"},
                "description": {"type": "string"},
                "active": {"type": "boolean", "default": "True"},
                "data": {
                    "type": "string",
                    "format": "json",
                    "description": "Free JSON field to add metadata",
                },
                "scene_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Target scene",
                },
                "target_asset_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Use when instantiating an asset in an asset is required",
                },
            },
        },
        "AssetType": {
            "type": "object",
            "properties": {"name": {"type": "string"}},
        },
        "AttachmentFile": {
            "type": "object",
            "properties": {
                "name": {
                    "type": "string",
                    "description": "Name of attachment file",
                },
                "size": {
                    "type": "integer",
                    "description": "Size of attachment file",
                },
                "extension": {
                    "type": "string",
                    "description": "Extension of attachment file",
                },
                "mimetype": {"type": "string"},
                "comment_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Comment to which the file is attached",
                },
            },
        },
        "BuildJob": {
            "type": "object",
            "properties": {
                "status": {
                    "type": "string",
                    "description": "Status of build job (running, failed, succeeded)",
                },
                "job_type": {
                    "type": "string",
                    "description": "Type of build job (archive, movie)",
                },
                "ended_at": {"type": "string", "format": "date-time"},
                "playlist_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Playlist ID",
                },
            },
        },
        "Comment": {
            "type": "object",
            "properties": {
                "shotgun_id": {
                    "type": "integer",
                    "description": "Used for synchronization with a Shotgun instance",
                },
                "object_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Unique ID of the commented model instance",
                },
                "object_type": {
                    "type": "string",
                    "description": "Model type of the comment model instance",
                },
                "text": {"type": "string"},
                "data": {
                    "type": "string",
                    "format": "json",
                    "description": "Free JSON field to add metadata",
                },
                "replies": {
                    "type": "string",
                    "format": "json",
                    "default": "[]",
                },
                "checklist": {"type": "string", "format": "json"},
                "pinned": {"type": "boolean"},
                "task_status_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Task status attached to comment",
                },
                "person_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "The person who publishes the comment",
                },
                "preview_file_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "ID of preview file used as thumbnail",
                },
            },
        },
        "CustomAction": {
            "type": "object",
            "properties": {
                "name": {
                    "type": "string",
                    "description": "Name of custom action",
                },
                "url": {"type": "string"},
                "entity_type": {"type": "string", "default": "all"},
                "is_ajax": {
                    "type": "boolean",
                    "default": "False",
                    "description": "True if the custom action is ajax, False otherwise",
                },
            },
        },
        "DataImportError": {
            "type": "object",
            "properties": {
                "event_data": {
                    "type": "string",
                    "format": "json",
                    "description": "JSON field to add event data",
                },
                "source": {
                    "type": "array",
                    "items": {"type": "string", "enum": ["csv", "shotgun"]},
                },
            },
        },
        "DayOff": {
            "type": "object",
            "properties": {
                "date": {"type": "string", "format": "date"},
                "person_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "The person who will take the day off",
                },
            },
        },
        "Department": {
            "type": "object",
            "properties": {
                "name": {
                    "type": "string",
                    "description": "Name of department",
                },
                "color": {
                    "type": "string",
                    "description": "Color of department",
                },
                "archived": {
                    "type": "boolean",
                    "description": "True if the department is archived else False",
                },
            },
        },
        "DesktopLoginLog": {
            "type": "object",
            "properties": {
                "date": {"type": "string", "format": "date"},
                "person_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Person ID",
                },
            },
        },
        "Episode": {
            "type": "object",
            "properties": {
                "name": {"type": "string", "description": "Name of episode"},
                "code": {
                    "type": "string",
                    "description": "Utility field for the pipeline to identify the episode",
                },
                "description": {
                    "type": "string",
                    "description": "Episode brief",
                },
                "canceled": {
                    "type": "boolean",
                    "default": "False",
                    "description": "True if the episode has been delete one time, False otherwise",
                },
                "project_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Project ID",
                },
                "source_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Field uset to set the episode_id",
                },
                "preview_file_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "ID of preview file used as thumbnail",
                },
                "data": {
                    "type": "string",
                    "format": "json",
                    "description": "Free JSON field to add metadata",
                },
                "shotgun_id": {
                    "type": "integer",
                    "description": "Used for synchronization with a Shotgun instance",
                },
            },
        },
        "Event": {
            "type": "object",
            "properties": {
                "name": {"type": "string", "description": "Name of event"},
                "user_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "The user who made the action that emitted the event",
                },
                "project_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Project ID",
                },
                "data": {
                    "type": "string",
                    "format": "json",
                    "description": "Free JSON field to add metadata",
                },
            },
        },
        "FileStatus": {
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "color": {"type": "string"},
            },
        },
        "LoginLog": {
            "type": "object",
            "properties": {
                "origin": {"type": "string", "description": "web, script"},
                "ip_address": {
                    "type": "string",
                    "description": "IP address of device used to login",
                },
                "person_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Person ID",
                },
            },
        },
        "Metadata": {
            "type": "object",
            "properties": {
                "project_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "ID of project for which metadata are added",
                },
                "entity_type": {
                    "type": "string",
                    "description": "Asset or Shot",
                },
                "name": {
                    "type": "string",
                    "description": "Field name for GUI",
                },
                "field_name": {
                    "type": "string",
                    "description": "Technical field name",
                },
                "choices": {
                    "type": "string",
                    "format": "json",
                    "description": "Array of string that represents the available values for this metadate (this metatada is considered as a free field if this array is empty)",
                },
                "for_client": {"type": "boolean"},
            },
        },
        "Milestone": {
            "type": "object",
            "properties": {
                "data": {
                    "type": "string",
                    "format": "date",
                    "description": "Milestone date of production schedule",
                },
                "name": {"type": "string", "description": "Name of milestone"},
                "task_type_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Task type ID",
                },
                "project_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Project ID",
                },
            },
        },
        "News": {
            "type": "object",
            "properties": {
                "change": {"type": "boolean", "default": "False"},
                "author_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Person who wrote the comment",
                },
                "comment_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Posted comment ID",
                },
                "task_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Task ID",
                },
                "preview_file_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Preview file ID",
                },
            },
        },
        "Notification": {
            "type": "object",
            "properties": {
                "read": {
                    "type": "boolean",
                    "description": "True if user read it, False otherwise",
                },
                "change": {
                    "type": "boolean",
                    "description": "True if there is status change related to this status, False otherwise",
                },
                "person_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "The user to who the notification is aimed at",
                },
                "author_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Author of the event to notify",
                },
                "comment_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Comment related to the notification, if there is any",
                },
                "task_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Task related to the notification, if there is any",
                },
                "reply_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Reply related to notification",
                },
                "type": {
                    "type": "string",
                    "description": "Type of notification",
                },
            },
        },
        "Organisation": {
            "type": "object",
            "properties": {
                "name": {
                    "type": "string",
                    "description": "Name of organisaition",
                },
                "hours_by_day": {"type": "float"},
                "has_avatar": {
                    "type": "boolean",
                    "default": "False",
                    "description": "True if the organisation has an avatar, Flase otherwise",
                },
                "use_original_file_name": {
                    "type": "boolean",
                    "default": "False",
                    "description": "True if the organisation uses original file names, Flase otherwise",
                },
                "timesheets_locked": {
                    "type": "boolean",
                    "default": "False",
                    "description": "True if the organisation's timesheets are locked, False otherwise",
                },
                "hd_by_default": {"type": "boolean", "default": "False"},
                "chat_token_slack": {"type": "string"},
                "chat_webhook_mattermost": {"type": "string"},
                "chat_token_discord": {"type": "string"},
            },
        },
        "OutputFile": {
            "type": "object",
            "properties": {
                "shotgun_id": {
                    "type": "integer",
                    "description": "Used for synchronization with a Shotgun instance",
                },
                "name": {
                    "type": "string",
                    "description": "Name of output file",
                },
                "extension": {
                    "type": "string",
                    "description": "Extension of output file",
                },
                "description": {
                    "type": "string",
                    "description": "Output file brief",
                },
                "comment": {
                    "type": "string",
                    "description": "Comment on output file",
                },
                "revision": {
                    "type": "integer",
                    "description": "Revision number of output file",
                },
                "size": {
                    "type": "integer",
                    "description": "Size of output file",
                },
                "checksum": {
                    "type": "string",
                    "description": "Checksum of output file",
                },
                "source": {
                    "type": "string",
                    "description": "Created by a script, a webgui or a desktop gui",
                },
                "path": {
                    "type": "string",
                    "description": "File path on the production hard drive",
                },
                "representation": {
                    "type": "string",
                    "description": "Precise what kind of output it is (abc, jpgs, pngs, etc.)",
                },
                "nb_elements": {
                    "type": "integer",
                    "default": "1",
                    "description": "For image sequence",
                },
                "canceled": {
                    "type": "boolean",
                },
                "file_status_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "File status ID",
                },
                "entity_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Asset or Shot concerned by the output file",
                },
                "asset_instance_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Asset instance ID",
                },
                "output_type_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Type of output (geometry, cache, etc.)",
                },
                "task_type_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Task type related to this output file (modeling, animation, etc.)",
                },
                "person_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Author of the file",
                },
                "source_file_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Working file that led to create this output file",
                },
                "temporal_entity_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Shot, scene or sequence needed for output files related to an asset instance",
                },
                "data": {
                    "type": "string",
                    "format": "json",
                    "description": "Free JSON field to add metadata",
                },
            },
        },
        "OutputType": {
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "short_name": {"type": "string"},
            },
        },
        "Person": {
            "type": "object",
            "properties": {
                "first_name": {"type": "string"},
                "last_name": {"type": "string"},
                "email": {"type": "string", "description": "Serve as login"},
                "phone": {"type": "string"},
                "active": {
                    "type": "boolean",
                    "description": "True if the person is still in the studio, False otherwise",
                },
                "last_presence": {
                    "type": "string",
                    "format": "date",
                    "description": "Last time the person worked for the studio",
                },
                "password": {"type": "string", "format": "byte"},
                "desktop_login": {
                    "type": "string",
                    "description": "Login used on desktop",
                },
                "shotgun_id": {
                    "type": "integer",
                    "description": "Used for synchronization with a Shotgun instance",
                },
                "timezone": {"type": "string"},
                "locale": {"type": "string"},
                "data": {
                    "type": "string",
                    "format": "json",
                    "description": "Free JSON field to add metadata",
                },
                "role": {"type": "string", "default": "user"},
                "has_avatar": {
                    "type": "boolean",
                    "default": "False",
                    "description": "True if user has an avatar, Flase otherwise",
                },
                "notifications_enabled": {
                    "type": "boolean",
                },
                "notifications_slack_enabled": {
                    "type": "boolean",
                },
                "notifications_slack_userid": {
                    "type": "string",
                },
                "notifications_mattermost_enabled": {
                    "type": "boolean",
                },
                "notifications_mattermost_userid": {
                    "type": "string",
                },
                "notifications_discord_enabled": {
                    "type": "boolean",
                },
                "notifications_discord_userid": {
                    "type": "string",
                },
            },
        },
        "Playlist": {
            "type": "object",
            "properties": {
                "name": {"type": "string", "description": "Name of playlist"},
                "shots": {
                    "type": "string",
                    "format": "json",
                    "description": "JSON field describing shot and preview listed in",
                },
                "project_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Project ID",
                },
                "episode_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Episode ID",
                },
                "for_client": {"type": "boolean", "default": "False"},
                "for_entity": {"type": "string", "default": "shot"},
                "is_for_all": {"type": "boolean", "default": "False"},
                "task_type_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Task type ID",
                },
            },
        },
        "PreviewFile": {
            "type": "object",
            "properties": {
                "shotgun_id": {
                    "type": "integer",
                    "description": "Used for synchronization with a Shotgun instance",
                },
                "name": {
                    "type": "string",
                    "description": "Name of preview file",
                },
                "original_name": {
                    "type": "string",
                    "description": "Original name of preview file",
                },
                "revision": {
                    "type": "integer",
                    "default": "1",
                    "description": "Revision number of preview file",
                },
                "position": {
                    "type": "integer",
                    "default": "1",
                    "description": "Position of preview file",
                },
                "extension": {
                    "type": "string",
                    "description": "Extension of preview file",
                },
                "description": {
                    "type": "string",
                    "description": "Preview file brief",
                },
                "path": {
                    "type": "string",
                    "description": "File path on the production hard drive",
                },
                "source": {
                    "type": "string",
                    "description": "Created by a script, a webgui or a desktop gui",
                },
                "file_size": {
                    "type": "integer",
                    "default": "0",
                    "description": "Size of output file",
                },
                "comment": {
                    "type": "string",
                    "description": "Comment on output file",
                },
                "checksum": {
                    "type": "string",
                    "description": "Checksum of output file",
                },
                "representation": {
                    "type": "string",
                    "description": "Precise what kind of output it is (abc, jpgs, pngs, etc.)",
                },
                "nb_elements": {
                    "type": "integer",
                    "default": "1",
                    "description": "For image sequence",
                },
                "canceled": {
                    "type": "boolean",
                },
                "file_status_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "File status ID",
                },
                "entity_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Asset or Shot concerned by the output file",
                },
                "asset_instance_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Asset instance ID",
                },
                "output_type_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Type of output (geometry, cache, etc.)",
                },
                "task_type_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Task type related to this output file (modeling, animation, etc.)",
                },
                "person_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Author of the file",
                },
                "source_file_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Working file that led to create this output file",
                },
                "temporal_entity_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Shot, scene or sequence needed for output files related to an asset instance",
                },
                "data": {
                    "type": "string",
                    "format": "json",
                    "description": "Free JSON field to add metadata",
                },
            },
        },
        "Project": {
            "type": "object",
            "properties": {
                "name": {"type": "string", "description": "Name of project"},
                "code": {
                    "type": "string",
                    "description": "Utility field for the pipeline to identify the project",
                },
                "description": {
                    "type": "string",
                    "description": "Project brief",
                },
                "shotgun_id": {
                    "type": "integer",
                    "description": "Used for synchronization with a Shotgun instance",
                },
                "file_tree": {
                    "type": "string",
                    "format": "json",
                    "description": "Templates to use to build file paths",
                },
                "data": {
                    "type": "string",
                    "format": "json",
                    "description": "Free JSON field to add metadata",
                },
                "project_status_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Project status ID",
                },
                "has_avatar": {
                    "type": "boolean",
                    "default": "False",
                    "description": "True if the project has an avatar, False otherwise",
                },
                "fps": {"type": "string", "description": "Frames per second"},
                "ratio": {"type": "string"},
                "resolution": {"type": "string"},
                "production_type": {
                    "type": "string",
                    "description": "short, featurefilm or tvshow",
                    "default": "short",
                },
                "end_date": {"type": "string", "format": "date"},
                "start_date": {"type": "string", "format": "date"},
                "man_days": {
                    "type": "integer",
                    "description": "Estimated number of working days to complete the project",
                },
                "nb_episodes": {"type": "integer", "default": "0"},
                "episode_span": {"type": "integer", "default": "0"},
                "max_retakes": {"type": "integer", "default": "0"},
                "is_clients_isolated": {
                    "type": "boolean",
                    "default": "False",
                    "description": "True if the clients are isolated from the project, False otherwise",
                },
            },
        },
        "ProjectStatus": {
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "color": {"type": "string"},
            },
        },
        "ScheduleItem": {
            "type": "object",
            "properties": {
                "start_date": {"type": "string", "format": "date"},
                "end_date": {"type": "string", "format": "date"},
                "man_days": {"type": "string", "format": "date"},
                "project_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Project ID",
                },
                "task_type_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Task type ID",
                },
                "object_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Object ID",
                },
            },
        },
        "SearchFilter": {
            "type": "object",
            "properties": {
                "list_type": {"type": "string", "description": "Type of list"},
                "entity_type": {
                    "type": "string",
                    "description": "Type of entity",
                },
                "name": {
                    "type": "string",
                    "description": "Name of search filter",
                },
                "search_query": {"type": "string"},
                "person_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Person ID",
                },
                "project_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Project ID",
                },
            },
        },
        "Sequence": {
            "type": "object",
            "properties": {
                "name": {"type": "string", "description": "Name of sequence"},
                "code": {
                    "type": "string",
                    "description": "Utility field for the pipeline to identify the sequence",
                },
                "description": {
                    "type": "string",
                    "description": "Sequence brief",
                },
                "canceled": {
                    "type": "boolean",
                    "default": "False",
                    "description": "True if the sequence has been delete one time, False otherwise",
                },
                "project_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Project ID",
                },
                "parent_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Episode ID",
                },
                "source_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Field uset to set the episode_id",
                },
                "preview_file_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "ID of preview file used as thumbnail",
                },
                "data": {
                    "type": "string",
                    "format": "json",
                    "description": "Free JSON field to add metadata",
                },
                "shotgun_id": {
                    "type": "integer",
                    "description": "Used for synchronization with a Shotgun instance",
                },
            },
        },
        "Shot": {
            "type": "object",
            "properties": {
                "name": {"type": "string", "description": "Name of shot"},
                "code": {
                    "type": "string",
                    "description": "Utility field for the pipeline to identify the shot",
                },
                "description": {"type": "string", "description": "Shot brief"},
                "canceled": {
                    "type": "boolean",
                    "default": "False",
                    "description": "True if the shot has been delete one time, False otherwise",
                },
                "project_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Project ID",
                },
                "parent_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Episode ID",
                },
                "entity_type_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Shot type ID",
                },
                "source_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Field uset to set the episode_id",
                },
                "preview_file_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "ID of preview file used as thumbnail",
                },
                "data": {
                    "type": "string",
                    "format": "json",
                    "description": "Free JSON field to add metadata",
                },
                "shotgun_id": {
                    "type": "integer",
                    "description": "Used for synchronization with a Shotgun instance",
                },
            },
        },
        "Software": {
            "type": "object",
            "properties": {
                "name": {"type": "string", "description": "Name of software"},
                "short_name": {
                    "type": "string",
                    "description": "Short name of software",
                },
                "file_extension": {
                    "type": "string",
                    "description": "Main extension used for this software's files",
                },
                "secondary_extensions": {
                    "type": "string",
                    "format": "json",
                    "description": "Other extensions used for this software's files",
                },
            },
        },
        "StatusAutomation": {
            "type": "object",
            "properties": {
                "entity_type": {"type": "string", "default": "asset"},
                "in_task_type_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Task type ID",
                },
                "in_task_status_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Task status ID",
                },
                "out_field_type": {
                    "type": "string",
                    "description": "Field type (status, ready_for)",
                },
                "out_task_type_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Task type ID",
                },
                "out_task_status_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Task status ID",
                },
            },
        },
        "SubscriptionToNotifications": {
            "type": "object",
            "properties": {
                "person_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Person ID",
                },
                "task_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Task ID",
                },
                "entity_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Entity ID",
                },
                "task_type_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Task type ID",
                },
            },
        },
        "Studio": {
            "type": "object",
            "properties": {
                "name": {
                    "type": "string",
                    "description": "Name of studio",
                },
                "color": {
                    "type": "string",
                    "description": "Color of studio",
                },
                "archived": {
                    "type": "boolean",
                    "description": "True if the studio is archived else False",
                },
            },
        },
        "Task": {
            "type": "object",
            "properties": {
                "name": {"type": "string", "description": "Name of task"},
                "description": {"type": "string", "description": "Task brief"},
                "priority": {
                    "type": "integer",
                    "default": "0",
                    "description": "Priority of task",
                },
                "duration": {
                    "type": "float",
                    "default": "0",
                    "description": "Duration of task",
                },
                "estimation": {
                    "type": "float",
                    "default": "0",
                    "description": "Estimation of duration of task",
                },
                "completion_rate": {
                    "type": "integer",
                    "default": "0",
                    "description": "Completion rate of task",
                },
                "retake_count": {
                    "type": "integer",
                    "default": "0",
                    "description": "Retake count of task",
                },
                "sort_order": {
                    "type": "integer",
                    "default": "0",
                    "description": "Sort order of task",
                },
                "start_date": {"type": "string", "format": "date-time"},
                "due_date": {"type": "string", "format": "date-time"},
                "real_start_date": {"type": "string", "format": "date-time"},
                "end_date": {"type": "string", "format": "date-time"},
                "last_comment_date": {"type": "string", "format": "date-time"},
                "nb_assets_ready": {
                    "type": "integer",
                    "default": "0",
                    "description": "Number of assets ready",
                },
                "data": {
                    "type": "string",
                    "format": "json",
                    "description": "Free JSON field to add metadata",
                },
                "shotgun_id": {
                    "type": "integer",
                    "description": "Used for synchronization with a Shotgun instance",
                },
                "project_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Project ID",
                },
                "task_type_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Task type ID",
                },
                "task_status_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Task status ID",
                },
                "entity_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Entity ID",
                },
                "assigner_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Person ID",
                },
            },
        },
        "TaskStatus": {
            "type": "object",
            "properties": {
                "name": {
                    "type": "string",
                    "description": "Name of task status",
                },
                "short_name": {
                    "type": "string",
                    "description": "Short name of task status",
                },
                "color": {"type": "string"},
                "is_done": {
                    "type": "boolean",
                    "default": "False",
                    "description": "True if the task is done, False otherwise",
                },
                "is_artist_allowed": {
                    "type": "boolean",
                    "default": "True",
                    "description": "True if the artist is allowed, False otherwise",
                },
                "is_client_allowed": {
                    "type": "boolean",
                    "default": "True",
                    "description": "True if the client is allowed, False otherwise",
                },
                "is_retake": {
                    "type": "boolean",
                    "default": "False",
                    "description": "True if the task was retaken, False otherwise",
                },
                "is_feedback_request": {
                    "type": "boolean",
                    "default": "False",
                    "description": "True if feedback was requested, False otherwise",
                },
                "is_default": {
                    "type": "boolean",
                    "default": "False",
                    "description": "True if the task is default, False otherwise",
                },
                "shotgun_id": {
                    "type": "integer",
                    "description": "Used for synchronization with a Shotgun instance",
                },
            },
        },
        "TaskType": {
            "type": "object",
            "properties": {
                "name": {"type": "string", "description": "Name of task type"},
                "short_name": {
                    "type": "string",
                    "description": "Short name of task type",
                },
                "color": {"type": "string", "default": "#FFFFFF"},
                "priority": {
                    "type": "integer",
                    "default": "1",
                    "description": "Priority of task type",
                },
                "for_entity": {"type": "string", "default": "Asset"},
                "allow_timelog": {
                    "type": "boolean",
                    "default": "True",
                    "description": "True if timelog is allowed, False otherwise",
                },
                "shotgun_id": {
                    "type": "integer",
                    "description": "Used for synchronization with a Shotgun instance",
                },
                "department_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Department ID",
                },
            },
        },
        "TimeSpent": {
            "type": "object",
            "properties": {
                "duration": {"type": "float"},
                "date": {"type": "string", "format": "date"},
                "task_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Related task ID",
                },
                "person_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "The person who performed the working time",
                },
            },
        },
        "WorkingFile": {
            "type": "object",
            "properties": {
                "shotgun_id": {
                    "type": "integer",
                    "description": "Used for synchronization with a Shotgun instance",
                },
                "name": {
                    "type": "string",
                    "description": "Name of working file",
                },
                "description": {
                    "type": "string",
                    "description": "working file brief",
                },
                "comment": {
                    "type": "string",
                    "description": "Comment on working file",
                },
                "revision": {
                    "type": "integer",
                    "description": "Revision number of working file",
                },
                "size": {
                    "type": "integer",
                    "description": "Size of working file",
                },
                "checksum": {
                    "type": "string",
                    "description": "Checksum of working file",
                },
                "path": {
                    "type": "string",
                    "description": "File path on the production hard drive",
                },
                "data": {
                    "type": "string",
                    "format": "json",
                    "description": "Free JSON field to add metadata",
                },
                "task_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Task for which the working file is made",
                },
                "entity_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Entity for which the working is made",
                },
                "person_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Author of the file",
                },
                "software_id": {
                    "type": "string",
                    "format": "UUID",
                    "description": "Software used to build this working file",
                },
            },
        },
    },
}



from flask import Blueprint
from zou.app.utils.api import configure_api_from_blueprint

from zou.app.blueprints.previews.resources import (
    AttachmentThumbnailResource,
    CreatePreviewFilePictureResource,
    PreviewFileLowMovieResource,
    PreviewFileMovieResource,
    PreviewFileMovieDownloadResource,
    PreviewFileThumbnailResource,
    PreviewFileResource,
    PreviewFileDownloadResource,
    PreviewFileThumbnailSquareResource,
    PreviewFilePreviewResource,
    PreviewFileOriginalResource,
    PreviewFileTileResource,
    OrganisationThumbnailResource,
    CreateOrganisationThumbnailResource,
    ProjectThumbnailResource,
    CreateProjectThumbnailResource,
    PersonThumbnailResource,
    CreatePersonThumbnailResource,
    RunningPreviewFiles,
    SetMainPreviewResource,
    UpdateAnnotationsResource,
    UpdatePreviewPositionResource,
    ExtractFrameFromPreview,
    ExtractTileFromPreview,
    CreatePreviewBackgroundFileResource,
    PreviewBackgroundFileThumbnailResource,
    PreviewBackgroundFileResource,
)

routes = [
    ("/data/playlists/preview-files/running", RunningPreviewFiles),
    (
        "/pictures/preview-files/<instance_id>",
        CreatePreviewFilePictureResource,
    ),
    (
        "/movies/originals/preview-files/<instance_id>.mp4",
        PreviewFileMovieResource,
    ),
    (
        "/movies/originals/preview-files/<instance_id>/download",
        PreviewFileMovieDownloadResource,
    ),
    (
        "/movies/low/preview-files/<instance_id>.mp4",
        PreviewFileLowMovieResource,
    ),
    (
        "/pictures/thumbnails/preview-files/<instance_id>.png",
        PreviewFileThumbnailResource,
    ),
    (
        "/pictures/thumbnails/attachment-files/<attachment_file_id>.png",
        AttachmentThumbnailResource,
    ),
    (
        "/pictures/thumbnails-square/preview-files/<instance_id>.png",
        PreviewFileThumbnailSquareResource,
    ),
    (
        "/pictures/originals/preview-files/<instance_id>.png",
        PreviewFileOriginalResource,
    ),
    (
        "/pictures/originals/preview-files/<instance_id>.<extension>",
        PreviewFileResource,
    ),
    (
        "/pictures/originals/preview-files/<instance_id>/download",
        PreviewFileDownloadResource,
    ),
    (
        "/pictures/previews/preview-files/<instance_id>.png",
        PreviewFilePreviewResource,
    ),
    (
        "/movies/tiles/preview-files/<instance_id>.png",
        PreviewFileTileResource,
    ),
    (
        "/pictures/thumbnails/organisations/<instance_id>",
        CreateOrganisationThumbnailResource,
    ),
    (
        "/pictures/thumbnails/organisations/<instance_id>.png",
        OrganisationThumbnailResource,
    ),
    (
        "/pictures/thumbnails/persons/<instance_id>",
        CreatePersonThumbnailResource,
    ),
    (
        "/pictures/thumbnails/persons/<instance_id>.png",
        PersonThumbnailResource,
    ),
    (
        "/pictures/thumbnails/projects/<instance_id>",
        CreateProjectThumbnailResource,
    ),
    (
        "/pictures/thumbnails/projects/<instance_id>.png",
        ProjectThumbnailResource,
    ),
    (
        "/pictures/preview-background-files/<instance_id>",
        CreatePreviewBackgroundFileResource,
    ),
    (
        "/pictures/thumbnails/preview-background-files/<instance_id>.png",
        PreviewBackgroundFileThumbnailResource,
    ),
    (
        "/pictures/preview-background-files/<instance_id>.<extension>",
        PreviewBackgroundFileResource,
    ),
    (
        "/actions/preview-files/<preview_file_id>/set-main-preview",
        SetMainPreviewResource,
    ),
    (
        "/actions/preview-files/<preview_file_id>/extract-frame",
        ExtractFrameFromPreview,
    ),
    (
        "/actions/preview-files/<preview_file_id>/update-position",
        UpdatePreviewPositionResource,
    ),
    (
        "/actions/preview-files/<preview_file_id>/update-annotations",
        UpdateAnnotationsResource,
    ),
    (
        "/actions/preview-files/<preview_file_id>/extract-tile",
        ExtractTileFromPreview,
    ),
]
blueprint = Blueprint("thumbnails", "thumbnails")
api = configure_api_from_blueprint(blueprint, routes)



import datetime
import logging
import os
import sys
import time
import traceback
import requests

import gazu
import sqlalchemy

from flask_fs.backends.local import LocalBackend
from http.client import responses as http_responses
from threading import RLock
from multiprocessing.pool import ThreadPool as Pool

from zou.app.models.attachment_file import AttachmentFile
from zou.app.models.build_job import BuildJob
from zou.app.models.custom_action import CustomAction
from zou.app.models.comment import Comment
from zou.app.models.day_off import DayOff
from zou.app.models.department import Department
from zou.app.models.entity import Entity, EntityLink
from zou.app.models.entity_type import EntityType
from zou.app.models.event import ApiEvent
from zou.app.models.organisation import Organisation
from zou.app.models.metadata_descriptor import MetadataDescriptor
from zou.app.models.milestone import Milestone
from zou.app.models.news import News
from zou.app.models.notification import Notification
from zou.app.models.person import Person
from zou.app.models.playlist import Playlist
from zou.app.models.preview_background_file import PreviewBackgroundFile
from zou.app.models.preview_file import PreviewFile
from zou.app.models.project import Project
from zou.app.models.project_status import ProjectStatus
from zou.app.models.schedule_item import ScheduleItem
from zou.app.models.subscription import Subscription
from zou.app.models.search_filter import SearchFilter
from zou.app.models.search_filter_group import SearchFilterGroup
from zou.app.models.task import Task
from zou.app.models.task_status import TaskStatus
from zou.app.models.task_type import TaskType
from zou.app.models.time_spent import TimeSpent
from zou.app.models.studio import Studio

from zou.app.services import deletion_service, tasks_service, projects_service
from zou.app.stores import file_store
from zou.app.utils import events, date_helpers
from zou.app import app, config


logger = logging.getLogger()
logger.setLevel(os.environ.get("LOGLEVEL", "INFO").upper())
console_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
console_handler.setFormatter(formatter)
logger.addHandler(console_handler)
lock = RLock()


preview_folder = app.config["PREVIEW_FOLDER"]
local_picture = LocalBackend(
    "local", {"root": os.path.join(preview_folder, "pictures")}
)
local_movie = LocalBackend(
    "local", {"root": os.path.join(preview_folder, "movies")}
)
local_file = LocalBackend(
    "local", {"root": os.path.join(preview_folder, "files")}
)


event_name_model_map = {
    "attachment-file": AttachmentFile,
    "asset": Entity,
    "asset-type": EntityType,
    "build-job": BuildJob,
    "custom-action": CustomAction,
    "comment": Comment,
    "concept": Entity,
    "department": Department,
    "day-off": DayOff,
    "entity": Entity,
    "entity-link": EntityLink,
    "entity-type": EntityType,
    "episode": Entity,
    "event": ApiEvent,
    "organisation": Organisation,
    "metadata-descriptor": MetadataDescriptor,
    "milestone": Milestone,
    "news": News,
    "notification": Notification,
    "person": Person,
    "playlist": Playlist,
    "preview-background-file": PreviewBackgroundFile,
    "preview-file": PreviewFile,
    "project": Project,
    "project-status": ProjectStatus,
    "sequence": Entity,
    "shot": Entity,
    "schedule-item": ScheduleItem,
    "subscription": Subscription,
    "search-filter": SearchFilter,
    "search-filter-group": SearchFilterGroup,
    "studio": Studio,
    "task": Task,
    "task-status": TaskStatus,
    "task-type": TaskType,
    "time-spent": TimeSpent,
}

event_name_model_path_map = {
    "attachment-file": "attachment-files",
    "asset": "assets",
    "asset-type": "entity-types",
    "build-job": "build-jobs",
    "comment": "comments",
    "concept": "concepts",
    "custom-action": "custom-actions",
    "day-off": "day-offs",
    "department": "departments",
    "entity": "entities",
    "entity-link": "entity-links",
    "entity-type": "entity-types",
    "episode": "episodes",
    "event": "events",
    "metadata-descriptor": "metadata-descriptors",
    "milestone": "milestones",
    "news": "news",
    "notification": "notifications",
    "organisation": "organisations",
    "person": "persons",
    "playlist": "playlists",
    "preview-background-file": "preview-background-files",
    "preview-file": "preview-files",
    "project": "projects",
    "project-status": "project-status",
    "sequence": "sequences",
    "shot": "shots",
    "schedule-item": "schedule-items",
    "search-filter": "search-filters",
    "search-filter-group": "search-filter-groups",
    "subscription": "subscriptions",
    "studio": "studios",
    "task": "tasks",
    "task-status": "task-status",
    "task-type": "task-types",
    "time-spent": "time-spents",
}

project_events = [
    "episode",
    "sequence",
    "asset",
    "shot",
    "task",
    "preview-file",
    "time-spent",
    "playlist",
    "build-job",
    "comment",
    "concept",
    "attachment-file",
    "metadata-descriptor",
    "schedule-item",
    "subscription",
    "notification",
    "entity-link",
    "news",
    "milestone",
]

main_events = [
    "studio",
    "department",
    "task-type",
    "task-status",
    "custom-action",
    "organisation",
    "project-status",
    "asset-type",
    "person",
    "project",
]

thumbnail_events = [
    "organisation",
    "person",
    "project",
]

file_events = [
    "preview-background-file:add-file",
    "preview-file:add-file",
    "organisation:set-thumbnail",
    "person:set-thumbnail",
    "project:set-thumbnail",
]

special_events = [
    "preview-file:set-main",
    "shot:casting-update",
    "task:unassign",
    "task:assign",
]


def init(source, login, password, multithreaded=False, number_workers=30):
    """
    Set parameters for the client that will retrieve data from the source.
    """
    if multithreaded:
        adapter = requests.adapters.HTTPAdapter(
            pool_connections=number_workers,
            pool_maxsize=number_workers * 2,
            max_retries=3,
        )
        gazu.raw.default_client.session.mount(
            source,
            adapter,
        )

        if config.FS_BACKEND == "swift":
            for fs in [
                file_store.movies,
                file_store.pictures,
                file_store.files,
            ]:
                try:
                    fs.backend.conn.head_container(fs.backend.name)
                except:
                    pass
                http_con = fs.backend.conn.http_conn[1]
                url = http_con.parsed_url
                http_con.request_session.mount(
                    f"{url.scheme}://{url.netloc}",
                    adapter,
                )

    gazu.set_host(source)
    gazu.log_in(login, password)


def init_events_listener(source, event_source, login, password, logs_dir=None):
    """
    Set parameters for the client that will listen to events from the source.
    """
    gazu.set_event_host(event_source)
    gazu.set_host(source)
    gazu.log_in(login, password)
    if logs_dir is not None:
        set_logger(logs_dir)

    return gazu.events.init()


def set_logger(logs_dir):
    """
    Configure so that it logs results to a file stored in a given folder.
    """
    file_name = os.path.join(logs_dir, "zou_sync_changes.log")
    file_handler = logging.handlers.TimedRotatingFileHandler(
        file_name, when="D"
    )

    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)


def run_listeners(event_client):
    """
    Run event listener which will run all previously associated callbacks to
    """
    try:
        gazu.events.run_client(event_client)
    except KeyboardInterrupt:
        raise
    except Exception:
        logger.error("An error occured.", exc_info=1)
        run_listeners(event_client)


def run_main_data_sync(project=None):
    """
    Retrieve and import all cross-projects data from source instance.
    """
    for event in main_events:
        path = event_name_model_path_map[event]
        model = event_name_model_map[event]
        sync_entries(path, model, project=project)


def run_project_data_sync(project=None):
    """
    Retrieve and import all data related to projects from source instance.
    """
    if project:
        projects = [gazu.project.get_project_by_name(project)]
    else:
        projects = gazu.project.all_open_projects()
    for project in projects:
        logger.info("Syncing %s..." % project["name"])
        for event in project_events:
            logger.info("Syncing %ss..." % event)
            path = event_name_model_path_map[event]
            model = event_name_model_map[event]
            sync_project_entries(project, path, model)
        sync_entity_thumbnails(project, "assets")
        sync_entity_thumbnails(project, "shots")
        sync_entity_thumbnails(project, "concepts")
        logger.info("Sync of %s complete." % project["name"])


def run_other_sync(project=None, with_events=False):
    """
    Retrieve and import all search filters and events from source instance.
    """
    sync_entries("search-filter-groups", SearchFilterGroup, project=project)
    sync_entries("search-filters", SearchFilter, project=project)
    sync_entries("day-offs", DayOff, project=project)
    if with_events:
        sync_entries("events", ApiEvent, project=project)


def run_last_events_sync(minutes=0, limit=300):
    """
    Retrieve last events from source instance and import related data and
    action.
    """
    path = "events/last?limit=%s" % limit
    if minutes > 0:
        now = date_helpers.get_utc_now_datetime()
        min_before = now - datetime.timedelta(minutes=minutes)
        after = min_before.strftime("%Y-%m-%dT%H:%M:%S")
        path += "&before=%s" % now.strftime("%Y-%m-%dT%H:%M:%S")
        path += "&after=%s" % after
    events = gazu.client.fetch_all(path)
    events.reverse()
    for event in events:
        event_name = event["name"].split(":")[0]
        if event_name in event_name_model_map:
            try:
                sync_event(event)
            except Exception:
                pass


def run_last_events_files(minutes=0, limit=50):
    """
    Retrieve last events from source instance and import related data and
    action.
    """
    path = "events/last?only_files=true&limit=%s" % limit
    if minutes > 0:
        now = date_helpers.get_utc_now_datetime()
        min_before = now - datetime.timedelta(minutes=minutes)
        after = min_before.strftime("%Y-%m-%dT%H:%M:%S")
        path += "&before=%s" % now.strftime("%Y-%m-%dT%H:%M:%S")
        path += "&after=%s" % after
    events = gazu.client.fetch_all(path)
    events.reverse()
    for event in events:
        event_name = event["name"].split(":")[0]
        if event_name == "preview-file":
            preview_file = PreviewFile.get(event["data"]["preview_file_id"])
            if preview_file is not None:
                download_preview_from_another_instance(preview_file)
        elif event_name in ["preview-background-file"]:
            preview_background_file = PreviewBackgroundFile.get(
                event["data"]["preview_background_file_id"]
            )
            if preview_background_file is not None:
                download_preview_background_from_another_instance(
                    preview_background_file
                )
        else:
            download_thumbnail_from_another_instance(
                event_name, event["data"]["%s_id" % event_name]
            )


def sync_event(event):
    """
    From information given by an event, retrieve related data and apply it.
    """
    event_name = event["name"]
    [event_name, action] = event_name.split(":")

    model = event_name_model_map[event_name]
    path = event_name_model_path_map[event_name]

    if event_name == "metadata-descriptor":  # Backward compatibility
        if "metadata_descriptor_id" not in event["data"]:
            event_name = "descriptor"
    instance_id = event["data"]["%s_id" % event_name.replace("-", "_")]

    if action in ["update", "new"]:
        instance = gazu.client.fetch_one(path, instance_id)
        model.create_from_import(instance)
    elif action in ["delete"]:
        model.delete_from_import(instance_id)


def sync_entries(model_name, model, project=None):
    """
    Retrieve cross-projects data from source instance.
    """
    instances = []

    page = 1
    init = True
    results = {"nb_pages": 2}
    params = {
        "relations": "true",
    }
    if model_name == "persons":
        params["with_pass_hash"] = "true"
    if project is not None and model_name in [
        "projects",
        "search-filters",
        "search-filter-groups",
    ]:
        project = gazu.project.get_project_by_name(project)
        if model_name == "projects":
            params = {"id": project["id"]}
        elif model_name in ["search-filters", "search-filter-groups"]:
            params = {"project_id": project["id"]}
    while init or results["nb_pages"] >= page:
        params["page"] = page
        results = gazu.client.fetch_all(model_name, params=params)
        if model_name == "task-status" and results["data"]:
            results["data"] = [
                r for r in results["data"] if not r["for_concept"]
            ]
        instances += results["data"]
        page += 1
        init = False
        model.create_from_import_list(results["data"])

    logger.info("%s %s synced." % (len(instances), model_name))


def sync_project_entries(project, model_name, model):
    """
    Retrieve all project data from source instance.
    """
    instances = []
    page = 1
    init = True
    results = {"nb_pages": 2}
    result_length = 1
    if model_name not in [
        "tasks",
        "comments",
        "news",
        "notifications",
        "playlists",
        "preview-files",
    ]:  # not much data we retrieve all in a single request.
        path = "projects/%s/%s" % (project["id"], model_name)
        results = gazu.client.fetch_all(path)
        instances += results
        try:
            model.create_from_import_list(instances)
        except sqlalchemy.exc.IntegrityError:
            logger.error("An error occured", exc_info=1)

    elif model_name == "news":
        while init or result_length > 0:
            path = "projects/%s/%s?page=%d" % (project["id"], model_name, page)
            results = gazu.client.fetch_all(path)["data"]
            instances += results
            try:
                model.create_from_import_list(results)
            except sqlalchemy.exc.IntegrityError:
                logger.error("An error occured", exc_info=1)
            result_length = len(results)
            page += 1
            init = False

    else:  # Lot of data, we retrieve all through paginated requests.
        while init or results["nb_pages"] >= page:
            path = "projects/%s/%s?page=%d" % (project["id"], model_name, page)
            if model_name == "playlists":
                path = "projects/%s/playlists/all?page=%d" % (
                    project["id"],
                    page,
                )
            results = gazu.client.fetch_all(path)
            instances += results["data"]
            try:
                model.create_from_import_list(results["data"])
            except sqlalchemy.exc.IntegrityError:
                logger.error("An error occured", exc_info=1)
            page += 1
            init = False
    logger.info("    %s %s synced." % (len(instances), model_name))


def sync_entity_thumbnails(project, model_name):
    """
    Once every preview files and entities has been imported, this function
    allows you to import project entities again to set thumbnails id (link to
    a preview file) for all entities.
    """
    results = gazu.client.fetch_all(
        "projects/%s/%s" % (project["id"], model_name)
    )
    total = 0
    for result in results:
        if result.get("preview_file_id") is not None:
            entity = Entity.get(result["id"])
            try:
                entity.update(
                    {
                        "preview_file_id": result["preview_file_id"],
                        "updated_at": result["updated_at"],
                    }
                )
                total += 1
            except sqlalchemy.exc.IntegrityError:
                logger.error("An error occured", exc_info=1)
    logger.info("    %s %s thumbnails synced." % (total, model_name))


def add_main_sync_listeners(event_client):
    """
    Add listeners to manage CRUD events related to general data.
    """
    for event in main_events:
        path = event_name_model_path_map[event]
        model = event_name_model_map[event]
        add_sync_listeners(event_client, path, event, model)


def add_project_sync_listeners(event_client):
    """
    Add listeners to manage CRUD events related to open projects data.
    """
    for event in project_events:
        path = event_name_model_path_map[event]
        model = event_name_model_map[event]
        add_sync_listeners(event_client, path, event, model)


def add_special_sync_listeners(event_client):
    """
    Add listeners to forward all non CRUD events to local event broadcaster.
    """
    for event in special_events:
        gazu.events.add_listener(event_client, event, forward_event(event))


def add_sync_listeners(event_client, model_name, event_name, model):
    """
    Add Create, Update and Delete event listeners for givent model name to given
    event client.
    """
    gazu.events.add_listener(
        event_client,
        "%s:new" % event_name,
        create_entry(model_name, event_name, model, "new"),
    )
    gazu.events.add_listener(
        event_client,
        "%s:update" % event_name,
        create_entry(model_name, event_name, model, "update"),
    )
    gazu.events.add_listener(
        event_client,
        "%s:delete" % event_name,
        delete_entry(model_name, event_name, model),
    )


def create_entry(model_name, event_name, model, event_type):
    """
    Generate a function that creates a model each time a related creation event
    is retrieved. If it's an update event, it updates the model related to the
    event. Data are retrived through the HTTP client.
    It's useful to generate callbacks for event listener.
    """

    def create(data):
        if data.get("sync", False):
            return
        model_id_field_name = event_name.replace("-", "_") + "_id"
        model_id = data[model_id_field_name]
        try:
            instance = gazu.client.fetch_one(model_name, model_id)
            model.create_from_import(instance)
            forward_base_event(event_name, event_type, data)
            if event_type == "new":
                logger.info("Creation: %s %s" % (event_name, model_id))
            else:
                logger.info("Update: %s %s" % (event_name, model_id))
        except gazu.exception.RouteNotFoundException as e:
            logger.error("Route not found: %s" % e)
            logger.error("Fail %s created/updated %s" % (event_name, model_id))

    return create


def delete_entry(model_name, event_name, model):
    """
    Generate a function that delete a model each time a related deletion event
    is retrieved.
    It's useful to generate callbacks for event listener.
    """

    def delete(data):
        if data.get("sync", False):
            return
        model_id = data[event_name.replace("-", "_") + "_id"]
        if event_name == "comment":
            comment = deletion_service.remove_comment(model_id)
            tasks_service.reset_task_data(comment["object_id"])
        else:
            model.delete_all_by(id=model_id)
        forward_base_event(event_name, "delete", data)
        logger.info("Deletion: %s %s" % (model_name, model_id))

    return delete


def forward_event(event_name):
    """
    Generate a function that takes data in argument and that forwards it as
    given event name to the local event brodcaster.
    It's useful to generate callbacks for event listener.
    """

    def forward(data):
        if not data.get("sync", False):
            data["sync"] = True
            logger.info("Forward event: %s" % event_name)
            project_id = data.get("project_id", None)
            events.emit(event_name, data, persist=False, project_id=project_id)

    return forward


def forward_base_event(event_name, event_type, data):
    """
    Forward given event to current instance event queue.
    """
    full_event_name = "%s:%s" % (event_name, event_type)
    data["sync"] = True
    logger.info("Forward event: %s" % full_event_name)
    project_id = data.get("project_id", None)
    events.emit(full_event_name, data, project_id=project_id)


def add_file_listeners(event_client):
    """
    Add new preview event listener.
    """
    gazu.events.add_listener(
        event_client, "preview-file:add-file", retrieve_preview_file
    )
    gazu.events.add_listener(
        event_client,
        "preview-background-file:add-file",
        retrieve_preview_background_file,
    )
    for model_name in thumbnail_events:
        gazu.events.add_listener(
            event_client,
            "%s:set-thumbnail" % model_name,
            get_retrieve_thumbnail(model_name),
        )


def retrieve_preview_file(data):
    if data.get("sync", False):
        return
    try:
        preview_file_id = data["preview_file_id"]
        preview_file = PreviewFile.get(preview_file_id)
        download_preview_from_another_instance(preview_file)
        forward_event({"name": "preview-file:add-file", "data": data})
        logger.info(
            "Preview file and related downloaded: %s" % preview_file_id
        )
    except gazu.exception.RouteNotFoundException as e:
        logger.error("Route not found: %s" % e)
        logger.error("Fail to dowonload preview file: %s" % (preview_file_id))


def retrieve_preview_background_file(data):
    if data.get("sync", False):
        return
    try:
        preview_background_file_id = data["preview_background_file_id"]
        preview_background_file = PreviewBackgroundFile.get(
            preview_background_file_id
        )
        download_preview_background_from_another_instance(
            preview_background_file
        )
        forward_event(
            {"name": "preview-background-file:add-file", "data": data}
        )
        logger.info(
            "Preview background file and related downloaded: %s"
            % preview_background_file_id
        )
    except gazu.exception.RouteNotFoundException as e:
        logger.error("Route not found: %s" % e)
        logger.error(
            "Fail to dowonload preview background file: %s"
            % (preview_background_file_id)
        )


def get_retrieve_thumbnail(model_name):
    def retrieve_thumbnail(data):
        if data.get("sync", False):
            return
        try:
            instance_id = data["preview_file_id"]
            download_thumbnail_from_another_instance(model_name, instance_id)
            forward_event(
                {"name": "%s:set-thumbnail" % model_name, "data": data}
            )
            logger.info(
                "Thumbnail downloaded: %s %s" % (model_name, instance_id)
            )
        except gazu.exception.RouteNotFoundException as e:
            logger.error("Route not found: %s" % e)
            logger.error(
                "Fail to dowonload thunbnail: %s %s"
                % (model_name, instance_id)
            )

    return retrieve_thumbnail


def download_entity_thumbnails_from_storage():
    """
    Download all thumbnail files for non preview entries from object storage
    and store them locally.
    """
    for project in Project.query.all():
        download_entity_thumbnail(project)
    for organisation in Organisation.query.all():
        download_entity_thumbnail(organisation)
    for person in Person.query.all():
        download_entity_thumbnail(person)


def download_preview_files_from_storage():
    """
    Download all thumbnail and original files for preview entries from object
    storage and store them locally.
    """
    for preview_file in PreviewFile.query.all():
        download_preview(preview_file)


def download_entity_thumbnail(entity):
    """
    Download thumbnail file for given entity from object storage and store it
    locally.
    """
    local = LocalBackend(
        "local", {"root": os.path.join(preview_folder, "pictures")}
    )

    file_path = local.path("thumbnails-" + str(entity.id))
    dirname = os.path.dirname(file_path)
    if entity.has_avatar:
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        with open(file_path, "wb") as tmp_file:
            for chunk in file_store.open_picture("thumbnails", str(entity.id)):
                tmp_file.write(chunk)


def download_file(file_path, prefix, dl_func, preview_file_id):
    """
    Download preview file for given preview from object storage and store it
    locally.
    """
    dirname = os.path.dirname(file_path)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    try:
        with open(file_path, "wb") as tmp_file:
            for chunk in dl_func(prefix, preview_file_id):
                tmp_file.write(chunk)
        logger.info("%s downloaded" % file_path)
    except BaseException:
        pass


def download_preview(preview_file):
    """
    Download all files link to preview file entry: orginal file and variants.
    """
    logger.info(
        "download preview %s (%s)" % (preview_file.id, preview_file.extension)
    )
    is_movie = preview_file.extension == "mp4"
    is_picture = preview_file.extension == "png"
    is_file = not is_movie and not is_picture

    preview_file_id = str(preview_file.id)
    file_key = "previews-%s" % preview_file_id
    if is_file:
        file_path = local_file.path(file_key)
        dl_func = file_store.open_picture
    elif is_movie:
        file_path = local_movie.path(file_key)
        dl_func = file_store.open_movie
    else:
        file_path = local_picture.path(file_key)
        dl_func = file_store.open_file

    if is_movie or is_picture:
        for prefix in ["thumbnails", "thumbnails-square", "original"]:
            pic_dl_func = file_store.open_picture
            pic_file_path = local_picture.path(
                "%s-%s" % (prefix, str(preview_file.id))
            )
            download_file(pic_file_path, prefix, pic_dl_func, preview_file_id)

    download_file(file_path, "previews", dl_func, preview_file_id)


def write_multithread_dict_errors(dict_errors, prefix, id, error):
    """
    Write a value in a dictionnary in a thread safe way.
    """
    with lock:
        if prefix not in dict_errors:
            dict_errors[prefix] = {}
        dict_errors[prefix][id] = error


def download_files_from_another_instance(
    project=None,
    multithreaded=False,
    number_workers=30,
    number_attemps=3,
    force_resync=False,
):
    """
    Download all files from source instance.
    """
    pool = None
    if multithreaded:
        pool = Pool(number_workers)

    dict_errors = {}

    download_thumbnails_from_another_instance(
        "person",
        pool=pool,
        number_attemps=number_attemps,
        force=force_resync,
        dict_errors=dict_errors,
    )
    download_thumbnails_from_another_instance(
        "organisation",
        pool=pool,
        number_attemps=number_attemps,
        force=force_resync,
        dict_errors=dict_errors,
    )
    download_thumbnails_from_another_instance(
        "project",
        project=project,
        pool=pool,
        number_attemps=number_attemps,
        force=force_resync,
        dict_errors=dict_errors,
    )
    download_preview_files_from_another_instance(
        project=project,
        pool=pool,
        number_attemps=number_attemps,
        force=force_resync,
        dict_errors=dict_errors,
    )
    download_preview_background_files_from_another_instance(
        project=project,
        pool=pool,
        number_attemps=number_attemps,
        force=force_resync,
        dict_errors=dict_errors,
    )
    download_attachment_files_from_another_instance(
        project=project,
        pool=pool,
        number_attemps=number_attemps,
        force=force_resync,
        dict_errors=dict_errors,
    )

    if pool is not None:
        pool.close()
        pool.join()

    return dict_errors


def download_thumbnails_from_another_instance(
    model_name,
    project=None,
    pool=None,
    number_attemps=3,
    force=False,
    dict_errors={},
):
    """
    Download all thumbnails from source instance for given model.
    """
    model = event_name_model_map[model_name]

    if project is None:
        instances = model.query
    else:
        project = gazu.project.get_project_by_name(project)
        instances = model.query.filter_by(id=project.get("id"))

    number_of_thumbnails = instances.count()
    logger.info(
        f"Downloading {model_name} thumbnails ({number_of_thumbnails})..."
    )
    for i, instance in enumerate(instances):
        if instance.has_avatar:
            args = (
                model_name,
                instance.id,
                number_attemps,
                i + 1,
                number_of_thumbnails,
                force,
                dict_errors,
            )
            if pool is None:
                download_thumbnail_from_another_instance(*args)
            else:
                pool.apply_async(
                    download_thumbnail_from_another_instance,
                    args,
                )


def download_thumbnail_from_another_instance(
    model_name,
    model_id,
    number_attemps=3,
    index=0,
    total=0,
    force=False,
    dict_errors={},
):
    """
    Download into the local storage the thumbnail for a given model instance.
    """
    file_path = f"/tmp/thumbnails-{model_id}.png"
    path = f"/pictures/thumbnails/{model_name}s/{model_id}.png"
    download_file_from_another_instance(
        path,
        file_path,
        file_store.exists_picture,
        file_store.add_picture,
        "thumbnails",
        model_id,
        number_attemps,
        force,
        dict_errors,
    )
    logger.info(
        f"{index:0{len(str(total))}}/{total} Thumbnail {model_name} file {model_id} processed."
    )


def download_preview_files_from_another_instance(
    project=None, pool=None, number_attemps=3, force=False, dict_errors={}
):
    """
    Download all preview files and related (thumbnails and low def included).
    """
    if project:
        project_dict = gazu.project.get_project_by_name(project)
        preview_files = PreviewFile.query.join(Task).filter(
            Task.project_id == project_dict["id"]
        )
    else:
        preview_files = PreviewFile.query

    number_of_preview_files = preview_files.count()
    logger.info(f"Downloading preview files ({number_of_preview_files})...")
    for i, preview_file in enumerate(preview_files):
        args = (
            preview_file,
            number_attemps,
            force,
            i + 1,
            number_of_preview_files,
            dict_errors,
        )
        if pool is None:
            download_preview_from_another_instance(*args)
        else:
            pool.apply_async(
                download_preview_from_another_instance,
                args,
            )


def download_preview_background_files_from_another_instance(
    project=None, pool=None, number_attemps=3, force=False, dict_errors={}
):
    """
    Download all preview background files and related.
    """
    if project:
        project_dict = gazu.project.get_project_by_name(project)
        project = projects_service.get_project_raw(project_dict["id"])
        preview_background_files = project.preview_background_files
        number_of_preview_background_files = len(preview_background_files)
    else:
        preview_background_files = PreviewBackgroundFile.query
        number_of_preview_background_files = preview_background_files.count()
    logger.info(
        f"Downloading preview background files ({number_of_preview_background_files})..."
    )
    for i, preview_background_file in enumerate(preview_background_files):
        args = (
            preview_background_file,
            number_attemps,
            force,
            i + 1,
            number_of_preview_background_files,
            dict_errors,
        )
        if pool is None:
            download_preview_background_from_another_instance(*args)
        else:
            pool.apply_async(
                download_preview_background_from_another_instance, args
            )


def download_preview_from_another_instance(
    preview_file,
    number_attemps=3,
    force=False,
    index=0,
    total=0,
    dict_errors={},
):
    """
    Download all files link to preview file entry: orginal file and variants.
    """
    is_movie = preview_file.extension == "mp4"
    is_picture = preview_file.extension == "png"
    is_file = not is_movie and not is_picture
    preview_file_id = str(preview_file.id)

    file_tree = {}
    if is_movie:
        file_tree[f"/movies/originals/preview-files/{preview_file_id}.mp4"] = {
            "prefix": "previews",
            "exist_func": file_store.exists_movie,
            "save_func": file_store.add_movie,
        }
        file_tree[f"/movies/low/preview-files/{preview_file_id}.mp4"] = {
            "prefix": "lowdef",
            "exist_func": file_store.exists_movie,
            "save_func": file_store.add_movie,
        }
        file_tree[f"/movies/tiles/preview-files/{preview_file_id}.png"] = {
            "prefix": "tiles",
            "exist_func": file_store.exists_picture,
            "save_func": file_store.add_picture,
        }
    if not is_file:
        file_tree[
            f"/pictures/thumbnails/preview-files/{preview_file_id}.png"
        ] = {
            "prefix": "thumbnails",
            "exist_func": file_store.exists_picture,
            "save_func": file_store.add_picture,
        }
        file_tree[
            f"/pictures/thumbnails-square/preview-files/{preview_file_id}.png"
        ] = {
            "prefix": "thumbnails-square",
            "exist_func": file_store.exists_picture,
            "save_func": file_store.add_picture,
        }
        file_tree[
            f"/pictures/previews/preview-files/{preview_file_id}.png"
        ] = {
            "prefix": "previews",
            "exist_func": file_store.exists_picture,
            "save_func": file_store.add_picture,
        }
        file_tree[
            f"/pictures/originals/preview-files/{preview_file_id}.png"
        ] = {
            "prefix": "original",
            "exist_func": file_store.exists_picture,
            "save_func": file_store.add_picture,
        }
    else:
        file_tree[
            f"/pictures/originals/preview-files/{preview_file_id}.{preview_file.extension}"
        ] = {
            "prefix": "previews",
            "exist_func": file_store.exists_file,
            "save_func": file_store.add_file,
        }

    for path, prefix_func in file_tree.items():
        file_path = f"/tmp/{prefix_func['prefix']}-{preview_file_id}.{preview_file.extension}"
        download_file_from_another_instance(
            path,
            file_path,
            prefix_func["exist_func"],
            prefix_func["save_func"],
            prefix_func["prefix"],
            preview_file_id,
            number_attemps,
            force,
            dict_errors,
        )

    logger.info(
        f"{index:0{len(str(total))}}/{total} Preview file {preview_file_id} processed."
    )


def download_preview_background_from_another_instance(
    preview_background,
    number_attemps=3,
    force=False,
    index=0,
    total=0,
    dict_errors={},
):
    """
    Download all files link to preview background file entry.
    """
    preview_background_file_id = str(preview_background.id)
    for prefix in [
        "thumbnails",
        "preview-backgrounds",
    ]:
        extension = (
            "png" if prefix == "thumbnails" else preview_background.extension
        )
        if prefix == "preview-backgrounds":
            path = f"/pictures/preview-background-files/{preview_background_file_id}.{extension}"
        elif prefix == "thumbnails":
            path = f"/pictures/thumbnails/preview-background-files/{preview_background_file_id}.png"

        file_path = f"/tmp/{prefix}-{preview_background_file_id}.{extension}"
        download_file_from_another_instance(
            path,
            file_path,
            file_store.exists_picture,
            file_store.add_picture,
            prefix,
            preview_background_file_id,
            number_attemps,
            force,
            dict_errors,
        )
    logger.info(
        f"{index:0{len(str(total))}}/{total} Preview background file {preview_background_file_id} processed."
    )


def download_attachment_files_from_another_instance(
    project=None, pool=None, number_attemps=3, force=False, dict_errors={}
):
    if project:
        project_dict = gazu.project.get_project_by_name(project)
        attachment_files = (
            AttachmentFile.query.join(Comment)
            .join(Task, Comment.object_id == Task.id)
            .filter(Task.project_id == project_dict["id"])
        )
    else:
        attachment_files = AttachmentFile.query

    number_of_attachment_files = attachment_files.count()
    logger.info(
        f"Downloading attachment files ({number_of_attachment_files})..."
    )
    for i, attachment_file in enumerate(attachment_files):
        args = (
            attachment_file.present(),
            number_attemps,
            i + 1,
            number_of_attachment_files,
            force,
            dict_errors,
        )
        if pool is None:
            download_attachment_file_from_another_instance(*args)
        else:
            pool.apply_async(
                download_attachment_file_from_another_instance,
                args,
            )


def download_attachment_file_from_another_instance(
    attachment_file,
    number_attemps=3,
    index=0,
    total=0,
    force=False,
    dict_errors={},
):
    attachment_file_id = attachment_file["id"]
    extension = attachment_file["extension"]
    path = "/data/attachment-files/%s/file/%s" % (
        attachment_file_id,
        attachment_file["name"],
    )
    file_path = "/tmp/%s.%s" % (attachment_file_id, extension)
    download_file_from_another_instance(
        path,
        file_path,
        file_store.exists_file,
        file_store.add_file,
        "attachments",
        attachment_file_id,
        number_attemps,
        force,
        dict_errors,
    )
    logger.info(
        f"{index:0{len(str(total))}}/{total} Attachment file {attachment_file_id} processed."
    )


def download_file_from_another_instance(
    path,
    file_path,
    exist_func,
    save_func,
    prefix,
    id,
    number_attemps=3,
    force=False,
    dict_errors={},
):
    with app.app_context():
        if force or not exist_func(prefix, id):
            for attemps_count in range(0, number_attemps):
                if attemps_count > 0:
                    time.sleep(0.5)
                try:
                    response = gazu.client.download(path, file_path)
                    if response.status_code != 200:
                        e = gazu.exception.DownloadFileException(
                            f"{response.status_code} {http_responses[response.status_code]}."
                        )
                        e.status_code = response.status_code
                        raise e
                except Exception as e:
                    if attemps_count + 1 == number_attemps:
                        if isinstance(e, gazu.exception.DownloadFileException):
                            error = f"Download failed ({path}):\n{e}"
                        else:
                            error = f"Download failed ({path}):\n{traceback.format_exc()}"
                        logger.error(error)

                        if (
                            not isinstance(
                                e, gazu.exception.DownloadFileException
                            )
                            or e.status_code != 404
                        ):
                            write_multithread_dict_errors(
                                dict_errors,
                                prefix,
                                id,
                                error,
                            )
                    if os.path.exists(file_path):
                        os.remove(file_path)
                    continue
                try:
                    save_func(prefix, id, file_path)
                    break
                except Exception:
                    if attemps_count + 1 == number_attemps:
                        error = f"Upload failed ({path}):\n{traceback.format_exc()}"
                        write_multithread_dict_errors(
                            dict_errors,
                            prefix,
                            id,
                            error,
                        )
                finally:
                    os.remove(file_path)
    return path, file_path



from slugify import slugify
from zou.app.blueprints.source.csv.base import (
    BaseCsvProjectImportResource,
    RowException,
)

from zou.app.services import (
    assets_service,
    shots_service,
    tasks_service,
    persons_service,
)

from zou.app.utils import date_helpers


class TaskTypeEstimationsCsvImportResource(BaseCsvProjectImportResource):
    def post(self, project_id, task_type_id, episode_id=None):
        """
        Import the estimations of task-types for given project.
        ---
        tags:
          - Import
        consumes:
          - multipart/form-data
        parameters:
          - in: formData
            name: file
            type: file
            required: true
          - in: path
            name: project_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: task_type_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            201:
                description: Estimations imported
            400:
                description: Format error
        """
        return super().post(project_id, task_type_id, episode_id)

    def prepare_import(self, project_id, task_type_id, episode_id=None):
        self.organisation = persons_service.get_organisation()
        self.assets_map = {}
        self.shots_map = {}
        self.tasks_map = {}

        asset_types_map = {}
        asset_types = assets_service.get_asset_types_for_project(project_id)
        for asset_type in asset_types:
            asset_types_map[asset_type["id"]] = slugify(asset_type["name"])

        criterions_assets = {"project_id": project_id}
        if episode_id is not None and episode_id not in ["main", "all"]:
            criterions_assets["source_id"] = episode_id
        assets = assets_service.get_assets(criterions_assets)
        for asset in assets:
            key = "%s%s" % (
                asset_types_map[asset["entity_type_id"]],
                slugify(asset["name"]),
            )
            self.assets_map[key] = asset["id"]

        sequences_map = {}
        criterions_sequences = {"project_id": project_id}
        if episode_id is not None and episode_id not in ["main", "all"]:
            criterions_sequences["parent_id"] = episode_id
        sequences = shots_service.get_sequences(criterions_sequences)
        for sequence in sequences:
            sequences_map[sequence["id"]] = slugify(sequence["name"])

        shots = shots_service.get_shots({"project_id": project_id})
        for shot in shots:
            sequence_key = sequences_map.get(shot["parent_id"])
            if sequence_key is not None:
                key = "%s%s" % (sequence_key, slugify(shot["name"]))
                self.shots_map[key] = shot["id"]

        for task in tasks_service.get_tasks_for_project_and_task_type(
            project_id, task_type_id
        ):
            self.tasks_map[task["entity_id"]] = task["id"]

    def import_row(self, row, project_id, task_type_id, episode_id=None):
        key = slugify("%s%s" % (row["Parent"], row["Entity"]))

        if self.assets_map.get(key):
            entity_id = self.assets_map[key]
        elif self.shots_map.get(key):
            entity_id = self.shots_map[key]
        else:
            raise RowException(f"Entity {key} not found")

        new_data = {}

        if row.get("Estimation") not in [None, ""]:
            new_data["estimation"] = round(
                float(row["Estimation"])
                * self.organisation["hours_by_day"]
                * 60
            )

        if row.get("Start date") not in [None, ""]:
            new_data["start_date"] = date_helpers.get_date_from_string(
                row["Start date"]
            )

        if row.get("Due date") not in [None, ""]:
            new_data["due_date"] = date_helpers.get_date_from_string(
                row["Due date"]
            )
        elif new_data.get("start_date") and new_data.get("estimation"):
            new_data["due_date"] = date_helpers.add_business_days_to_date(
                new_data["start_date"], float(row["Estimation"]) - 1
            )

        tasks_service.update_task(self.tasks_map[entity_id], new_data)


class TaskTypeEstimationsEpisodeCsvImportResource(
    TaskTypeEstimationsCsvImportResource
):
    def post(self, project_id, task_type_id, episode_id):
        """
        Import the estimations of task-types for given episode of given project.
        ---
        tags:
          - Import
        consumes:
          - multipart/form-data
        parameters:
          - in: formData
            name: file
            type: file
            required: true
          - in: path
            name: project_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: task_type_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: episode_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            201:
                description: Estimations imported
            400:
                description: Format error
        """
        return super().post(project_id, task_type_id, episode_id)



from zou.app.models.attachment_file import AttachmentFile

from zou.app.blueprints.crud.base import BaseModelResource, BaseModelsResource

from zou.app.services import chats_service, tasks_service, user_service

from zou.app.utils.permissions import PermissionDenied


class AttachmentFilesResource(BaseModelsResource):
    def __init__(self):
        BaseModelsResource.__init__(self, AttachmentFile)


class AttachmentFileResource(BaseModelResource):
    def __init__(self):
        BaseModelResource.__init__(self, AttachmentFile)

    def check_read_permissions(self, instance):
        attachment_file = instance
        if attachment_file["comment_id"] is not None:
            comment = tasks_service.get_comment(attachment_file["comment_id"])
            task = tasks_service.get_task(comment["object_id"])
            user_service.check_project_access(task["project_id"])
            user_service.check_entity_access(task["entity_id"])
        elif attachment_file["chat_message_id"] is not None:
            message = chats_service.get_chat_message(
                attachment_file["chat_message_id"]
            )
            chat = chats_service.get_chat(message["chat_id"])
            print(chat)
            user_service.check_entity_access(chat["object_id"])
        else:
            raise PermissionDenied()
        return True



from zou.app.models.file_status import FileStatus
from zou.app.blueprints.crud.base import BaseModelResource, BaseModelsResource


class FileStatusesResource(BaseModelsResource):
    def __init__(self):
        BaseModelsResource.__init__(self, FileStatus)

    def check_read_permissions(self, options=None):
        return True


class FileStatusResource(BaseModelResource):
    def __init__(self):
        BaseModelResource.__init__(self, FileStatus)

    def check_read_permissions(self, instance):
        return True



"""empty message

Revision ID: 892b264937ec
Revises: df1834485f57
Create Date: 2019-05-08 13:57:15.653356

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
import sqlalchemy_utils
import uuid

# revision identifiers, used by Alembic.
revision = "892b264937ec"
down_revision = "df1834485f57"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "news",
        sa.Column(
            "id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.Column("change", sa.Boolean(), nullable=False),
        sa.Column(
            "person_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column(
            "author_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column(
            "comment_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=True,
        ),
        sa.Column(
            "task_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column(
            "preview_file_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["author_id"],
            ["person.id"],
        ),
        sa.ForeignKeyConstraint(
            ["comment_id"],
            ["comment.id"],
        ),
        sa.ForeignKeyConstraint(
            ["person_id"],
            ["person.id"],
        ),
        sa.ForeignKeyConstraint(
            ["preview_file_id"],
            ["task.id"],
        ),
        sa.ForeignKeyConstraint(
            ["task_id"],
            ["task.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_news_author_id"), "news", ["author_id"], unique=False
    )
    op.create_index(
        op.f("ix_news_comment_id"), "news", ["comment_id"], unique=False
    )
    op.create_index(
        op.f("ix_news_person_id"), "news", ["person_id"], unique=False
    )
    op.create_index(
        op.f("ix_news_preview_file_id"),
        "news",
        ["preview_file_id"],
        unique=False,
    )
    op.create_index(op.f("ix_news_task_id"), "news", ["task_id"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_news_task_id"), table_name="news")
    op.drop_index(op.f("ix_news_preview_file_id"), table_name="news")
    op.drop_index(op.f("ix_news_person_id"), table_name="news")
    op.drop_index(op.f("ix_news_comment_id"), table_name="news")
    op.drop_index(op.f("ix_news_author_id"), table_name="news")
    op.drop_table("news")
    # ### end Alembic commands ###



from flask import Blueprint
from zou.app.utils.api import configure_api_from_blueprint

from zou.app.blueprints.events.resources import (
    EventsResource,
    LoginLogsResource,
)

routes = [
    ("/data/events/last", EventsResource),
    ("/data/events/login-logs/last", LoginLogsResource),
]

blueprint = Blueprint("events", "events")
api = configure_api_from_blueprint(blueprint, routes)



from flask import current_app

from zou.app.models.preview_file import PreviewFile
from zou.app.models.person import Person


from zou.app.blueprints.source.shotgun.base import (
    BaseImportShotgunResource,
    ImportRemoveShotgunBaseResource,
)


class ImportShotgunVersionsResource(BaseImportShotgunResource):
    def __init__(self):
        BaseImportShotgunResource.__init__(self)

    def prepare_import(self):
        self.person_ids = Person.get_id_map()

    def filtered_entries(self):
        return (
            x for x in self.sg_entries if self.is_version_linked_to_task(x)
        )

    def is_version_linked_to_task(self, version):
        return version["sg_task"] is not None

    def extract_data(self, sg_version):
        data = {
            "name": sg_version["code"],
            "shotgun_id": sg_version["id"],
            "description": sg_version["description"],
            "source": "Shotgun",
        }

        if "user" in sg_version and sg_version["user"] is not None:
            data["person_id"] = self.person_ids.get(
                sg_version["user"]["id"], None
            )

        if sg_version["sg_task"] is not None:
            data["task_id"] = self.get_task_id(sg_version["sg_task"]["id"])

        if sg_version["sg_uploaded_movie"] is not None:
            data["uploaded_movie_url"] = sg_version["sg_uploaded_movie"]["url"]
            data["uploaded_movie_name"] = sg_version["sg_uploaded_movie"][
                "name"
            ]

        return data

    def import_entry(self, data):
        preview_file = PreviewFile.get_by(shotgun_id=data["shotgun_id"])
        if preview_file is None:
            preview_file = PreviewFile.get_by(
                name=data["name"], task_id=data["task_id"]
            )

        if preview_file is None:
            preview_file = PreviewFile(**data)
            preview_file.save()
            current_app.logger.info("PreviewFile created: %s" % preview_file)
        else:
            preview_file.update(data)
            current_app.logger.info("PreviewFile updated: %s" % preview_file)
        return preview_file


class ImportRemoveShotgunVersionResource(ImportRemoveShotgunBaseResource):
    def __init__(self):
        ImportRemoveShotgunBaseResource.__init__(self, PreviewFile)



from zou.app import app
from zou.app.indexer import indexing

from zou.app.services import (
    assets_service,
    persons_service,
    projects_service,
    shots_service,
    files_service,
)


def get_index(index_name):
    """
    Retrieve whoosh index from disk. It is required to perform any operations.
    """
    indexing.init()
    return indexing.get_index(index_name)


def get_asset_index():
    return get_index("assets")


def get_person_index():
    return get_index("persons")


def get_shot_index():
    return get_index("shots")


def chunks(lst, n):
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def reset_index():
    """
    Delete index and rebuild it by looping on all the assets listed in the
    database.
    """
    reset_asset_index()
    reset_person_index()
    reset_shot_index()


def reset_entry_index(
    index_name,
    get_entries,
    prepare_entry,
    searchable_fields=[],
    filterable_fields=[],
):
    """
    Clear and rebuild index for given parameters: folder name of the index,
    func to get entries to index, func to index a given entry.
    """
    index = indexing.create_index(
        index_name, searchable_fields, filterable_fields
    )
    indexing.clear_index(index_name)
    entries = get_entries()
    documents = []
    for chunk in chunks(entries, 3000):
        for entry in chunk:
            document = prepare_entry(entry)
            documents.append(document)
        indexing.index_documents(index, documents)
    print(len(entries), "%s indexed" % index_name)
    return entries


def reset_asset_index():
    reset_entry_index(
        "assets",
        assets_service.get_all_raw_assets,
        prepare_asset,
        searchable_fields=["name", "description", "metadatas"],
        filterable_fields=["project_id"],
    )


def reset_person_index():
    reset_entry_index(
        "persons",
        persons_service.get_all_raw_active_persons,
        prepare_person,
        searchable_fields=[
            "name",
        ],
    )


def reset_shot_index():
    reset_entry_index(
        "shots",
        shots_service.get_all_raw_shots,
        prepare_shot,
        searchable_fields=["name", "description", "metadatas"],
        filterable_fields=["project_id"],
    )


def search_assets(query, project_ids=[], limit=3, offset=0):
    """
    Perform a search on the index. The query is a simple string. The result is
    a list of assets with extra data like the project name and the asset type
    name (3 results maximum by default).
    """
    index = get_asset_index()
    assets = []

    results = indexing.search(
        index, query, project_ids, limit=limit, offset=offset
    )
    for asset_id, matched_terms in results:
        asset = assets_service.get_asset(asset_id)
        asset_type = assets_service.get_asset_type(asset["entity_type_id"])
        project = projects_service.get_project(asset["project_id"])
        asset["project_name"] = project["name"]
        asset["asset_type_name"] = asset_type["name"]
        if asset["preview_file_id"] is not None:
            preview_file = files_service.get_preview_file(
                asset["preview_file_id"]
            )
            asset["preview_file_extension"] = preview_file["extension"]
        else:
            asset["preview_file_extension"] = None
        asset["matched_terms"] = matched_terms
        assets.append(asset)
    return assets


def search_shots(query, project_ids=[], limit=3, offset=0):
    """
    Perform a search on the index. The query is a simple string. The result is
    a list of shots with extra data like the project name and the asset type
    name (3 results maximum by default).
    """
    index = get_shot_index()
    shots = []

    results = indexing.search(
        index, query, project_ids, limit=limit, offset=offset
    )

    for shot_id, matched_terms in results:
        shot = shots_service.get_shot(shot_id)
        sequence = shots_service.get_sequence(shot["parent_id"])
        project = projects_service.get_project(shot["project_id"])
        shot["project_name"] = project["name"]
        shot["sequence_name"] = sequence["name"]
        if shot["preview_file_id"] is not None:
            preview_file = files_service.get_preview_file(
                shot["preview_file_id"]
            )
            shot["preview_file_extension"] = preview_file["extension"]
        else:
            shot["preview_file_extension"] = None
        if projects_service.is_tv_show(project):
            episode_id = sequence.get("parent_id", None)
            if episode_id is not None:
                episode = shots_service.get_episode(episode_id)
                shot["episode"] = episode["name"]
                shot["episode_id"] = episode["id"]
        shot["matched_terms"] = matched_terms
        shots.append(shot)
    return shots


def search_persons(query, limit=3, offset=0):
    """
    Perform a search on the index. The query is a simple string. The result is
    a list of persons (3 results maximum by default).
    """
    index = get_person_index()
    persons = []
    results = indexing.search(index, query, limit=limit, offset=offset)
    for person_id, matched_terms in results:
        person = persons_service.get_person(person_id)
        person["matched_terms"] = matched_terms
        persons.append(person)
    return persons


def index_asset(asset):
    """
    Register asset into the index.
    """
    try:
        index = get_asset_index()
        document = prepare_asset(asset)
        indexing.index_document(index, document)
        return document
    except:
        app.logger.error("Indexer is not reachable, indexation failed.")
        return {}


def index_person(person):
    """
    Register person into the index.
    """
    try:
        index = get_person_index()
        document = prepare_person(person)
        indexing.index_document(index, document)
        return document
    except:
        app.logger.error("Indexer is not reachable, indexation failed.")
        return {}


def index_shot(shot):
    """
    Register shot into the index.
    """
    try:
        index = get_shot_index()
        document = prepare_shot(shot)
        indexing.index_document(index, document)
        return document
    except:
        app.logger.error("Indexer is not reachable, indexation failed.")
        return {}


def prepare_asset(asset):
    """
    Prepare a indexation document from given asset.
    """
    asset_serialized = asset.serialize()
    asset_type = assets_service.get_asset_type(
        asset_serialized["entity_type_id"]
    )
    metadatas = {}
    if asset_serialized["data"]:
        for k, v in asset_serialized["data"].items():
            metadatas[f"{k}"] = str(v)
    name = asset_serialized["name"]
    name = name + " " + name.replace("_", " ").replace("-", " ")
    name = asset_type["name"] + " " + name
    data = {
        "id": asset_serialized["id"],
        "name": name,
        "project_id": asset_serialized["project_id"],
        "episode_id": asset_serialized["source_id"] or "",
        "description": asset_serialized["description"],
        "metadatas": metadatas,
    }
    return data


def prepare_person(person, index=None):
    """
    Prepare a indexation document from given person.
    """
    if index is None:
        index = get_person_index()
    person_serialized = person.serialize()
    data = {
        "id": person_serialized["id"],
        "name": person_serialized["full_name"],
    }
    return data


def prepare_shot(shot, index=None):
    """
    Prepare a indexation document from given shot.
    """
    if index is None:
        index = get_shot_index()
    shot_serialized = shot.serialize()

    episode_id = ""
    episode = None
    sequence_id = shot_serialized.get("parent_id", "")
    if sequence_id != "":
        sequence = shots_service.get_sequence(sequence_id)
        episode_id = sequence.get("parent_id", "")
        if episode_id not in ["", "None", None]:
            episode = shots_service.get_episode(episode_id)

    shot_name = shot_serialized["name"]
    name = shot_name + " " + shot_name.replace("_", " ").replace("-", " ")
    if episode is not None:
        shot_name = f'{episode["name"]} {sequence["name"]} {name}'
    else:
        shot_name = f'{sequence["name"]} {name}'

    metadatas = {}
    if shot_serialized["data"]:
        for k, v in shot_serialized["data"].items():
            if k not in [
                "frame_in",
                "frame_out",
                "fps",
                "handle_in",
                "handle_out",
            ]:
                metadatas[f"{k}"] = str(v)

    data = {
        "id": shot_serialized["id"],
        "name": name,
        "project_id": shot_serialized["project_id"],
        "sequence_id": sequence_id,
        "episode_id": episode_id,
        "description": shot_serialized["description"] or "",
        "metadatas": metadatas,
    }
    return data


def remove_asset_index(asset_id):
    """
    Remove document matching given asset id from asset index.
    """
    try:
        return indexing.remove_document(get_asset_index(), asset_id)
    except:
        app.logger.error("Indexer is not reachable, indexation failed.")
        return {}


def remove_person_index(person_id):
    """
    Remove document matching given person id from person index.
    """
    try:
        return indexing.remove_document(get_person_index(), person_id)
    except:
        app.logger.error("Indexer is not reachable, indexation failed.")
        return {}


def remove_shot_index(shot_id):
    """
    Remove document matching given shot id from shot index.
    """
    try:
        return indexing.remove_document(get_shot_index(), shot_id)
    except:
        app.logger.error("Indexer is not reachable, indexation failed.")
        return {}



from flask import current_app

from zou.app.models.project import Project
from zou.app.models.entity import Entity

from zou.app.services import shots_service, persons_service

from zou.app.blueprints.source.shotgun.base import (
    BaseImportShotgunResource,
    ImportRemoveShotgunBaseResource,
)


class ImportShotgunScenesResource(BaseImportShotgunResource):
    def __init__(self):
        BaseImportShotgunResource.__init__(self)

    def prepare_import(self):
        self.scene_type = shots_service.get_shot_type()
        self.project_map = Project.get_id_map(field="name")
        self.current_user_id = persons_service.get_current_user()["id"]

    def extract_data(self, sg_scene):
        project_id = self.get_project(sg_scene, self.project_map)
        sequence_id = self.get_sequence(sg_scene)
        scene_type = shots_service.get_scene_type()

        data = {
            "name": sg_scene["code"],
            "shotgun_id": sg_scene["id"],
            "project_id": project_id,
            "entity_type_id": scene_type["id"],
            "parent_id": sequence_id,
        }
        return data

    def get_project(self, sg_scene, project_map):
        project_id = None
        if sg_scene["project"] is not None:
            project_id = project_map.get(sg_scene["project"]["name"], None)
        return project_id

    def get_sequence(self, sg_scene):
        sequence_id = None
        sequence_key = "sequence_sg_scenes_1_sequences"
        if (
            sequence_key in sg_scene
            and sg_scene[sequence_key] is not None
            and len(sg_scene[sequence_key]) > 0
        ):
            sequence_id = self.get_sequence_id(sg_scene[sequence_key][0]["id"])
        return sequence_id

    def import_entry(self, data):
        scene = Entity.get_by(
            shotgun_id=data["shotgun_id"],
            entity_type_id=shots_service.get_scene_type()["id"],
        )

        if scene is None:
            scene = Entity.create(**data, created_by=self.current_user_id)
            current_app.logger.info("Scene created: %s" % scene)

        else:
            scene.update(data)
            scene.save()
            current_app.logger.info("Scene updated: %s" % scene)

        return scene

    def filtered_entries(self):
        return self.sg_entries


class ImportRemoveShotgunSceneResource(ImportRemoveShotgunBaseResource):
    def __init__(self):
        ImportRemoveShotgunBaseResource.__init__(
            self, Entity, entity_type_id=shots_service.get_scene_type()["id"]
        )



"""add is clients isolated flag to projects

Revision ID: 4095103c7d01
Revises: 05ac7e8caa21
Create Date: 2022-06-22 15:09:56.585117

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "4095103c7d01"
down_revision = "05ac7e8caa21"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "project",
        sa.Column("is_clients_isolated", sa.Boolean(), nullable=True),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("project", "is_clients_isolated")
    # ### end Alembic commands ###



from flask_restful import reqparse
from flask import request

from zou.app.utils import date_helpers
from zou.app.services.exception import WrongParameterException


class ArgsMixin(object):
    """
    Helpers to retrieve parameters from GET or POST queries.
    """

    def get_args(self, descriptors, location=None):
        parser = reqparse.RequestParser()
        if location is None:
            location = ["values", "json"] if request.is_json else ["values"]

        for descriptor in descriptors:
            action = None
            data_type = str
            required = False
            default = None
            help = None

            if isinstance(descriptor, (list, tuple)):
                if len(descriptor) == 5:
                    (name, default, required, data_type, action) = descriptor
                elif len(descriptor) == 4:
                    (name, default, required, data_type) = descriptor
                elif len(descriptor) == 3:
                    (name, default, required) = descriptor
                elif len(descriptor) == 2:
                    (name, default) = descriptor
                elif len(descriptor) == 1:
                    (name) = descriptor
                else:
                    raise ValueError
            elif isinstance(descriptor, str):
                name = descriptor
            elif isinstance(descriptor, dict):
                name = descriptor.get("name")
                required = descriptor.get("required", required)
                default = descriptor.get("default", default)
                action = descriptor.get("action", action)
                data_type = descriptor.get("type", data_type)
                help = descriptor.get("help", help)

            parser.add_argument(
                name,
                required=required,
                default=default,
                action=action,
                type=data_type,
                help=help,
                location=location,
            )

        return parser.parse_args()

    def clear_empty_fields(self, data, ignored_fields=[]):
        """
        Remove fields set to None from data dict.
        """
        for key in list(data.keys()):
            if key not in ignored_fields and data[key] is None:
                del data[key]
        return data

    def get_page(self):
        """
        Returns page requested by the user as an integer.
        """
        options = request.args
        return int(options.get("page", "-1"))

    def get_limit(self):
        """
        Returns limit requested by the user as an integer.
        """
        options = request.args
        return int(options.get("limit", "-1"))

    def get_sort_by(self):
        """
        Returns sort by option value
        """
        return self.get_text_parameter("sort_by")

    def get_force(self):
        """
        Returns force parameter.
        """
        return self.get_bool_parameter("force")

    def get_relations(self):
        """
        Returns force parameter.
        """
        return self.get_bool_parameter("relations")

    def get_project_id(self):
        """
        Returns episode ID parameter.
        """
        return self.get_text_parameter("project_id")

    def get_episode_id(self):
        """
        Returns episode ID parameter.
        """
        return self.get_text_parameter("episode_id")

    def get_task_type_id(self):
        """
        Returns Task type ID parameter.
        """
        return self.get_text_parameter("task_type_id")

    def get_no_job(self):
        """
        Returns no_job parameter.
        """
        return self.get_bool_parameter("no_job")

    def get_text_parameter(self, field_name, default=None):
        """
        Returns text parameter value matching `field_name`.
        """
        options = request.args
        return options.get(field_name, default)

    def get_bool_parameter(self, field_name, default="false"):
        """
        Returns bool parameter value matching `field_name`.
        """
        options = request.args
        return options.get(field_name, default).lower() == "true"

    def get_date_parameter(self, field_name):
        """
        Returns date parameter value matching `field_name`.
        """
        self.parse_date_parameter(self.get_text_parameter(field_name))

    def parse_date_parameter(self, param):
        date = None
        if param is None:
            return date
        try:
            date = date_helpers.get_datetime_from_string(param)
        except Exception:
            try:
                date = date_helpers.get_date_from_string(param)
            except Exception:
                raise WrongParameterException(
                    "Wrong date format for before argument."
                    "Expected format: 2020-01-05T13:23:10 or 2020-01-05"
                )
        return date



from sqlalchemy.orm import relationship
from sqlalchemy_utils import UUIDType

from zou.app import db
from zou.app.models.serializer import SerializerMixin
from zou.app.models.base import BaseMixin

from sqlalchemy.dialects.postgresql import JSONB


class OutputFile(db.Model, BaseMixin, SerializerMixin):
    """
    Describe a file generated from a CG artist scene. It's the result of a
    publication.
    It is linked to a working file, an entity and a task type.
    """

    shotgun_id = db.Column(db.Integer())

    name = db.Column(db.String(250), nullable=False)
    canceled = db.Column(db.Boolean(), default=False, nullable=False)
    size = db.Column(db.Integer())
    checksum = db.Column(db.String(32))
    description = db.Column(db.Text())
    comment = db.Column(db.Text())
    extension = db.Column(db.String(10))
    revision = db.Column(db.Integer(), nullable=False)
    representation = db.Column(db.String(20), index=True)
    nb_elements = db.Column(db.Integer(), default=1)
    source = db.Column(db.String(40))
    path = db.Column(db.String(400))
    data = db.Column(JSONB)

    file_status_id = db.Column(
        UUIDType(binary=False),
        db.ForeignKey("file_status.id"),
        nullable=False,
        index=True,
    )
    entity_id = db.Column(
        UUIDType(binary=False), db.ForeignKey("entity.id"), index=True
    )
    asset_instance_id = db.Column(
        UUIDType(binary=False), db.ForeignKey("asset_instance.id"), index=True
    )
    output_type_id = db.Column(
        UUIDType(binary=False), db.ForeignKey("output_type.id"), index=True
    )
    task_type_id = db.Column(
        UUIDType(binary=False), db.ForeignKey("task_type.id"), index=True
    )
    person_id = db.Column(
        UUIDType(binary=False), db.ForeignKey("person.id"), index=True
    )
    source_file_id = db.Column(
        UUIDType(binary=False), db.ForeignKey("working_file.id"), index=True
    )
    source_file = relationship("WorkingFile", back_populates="outputs")
    temporal_entity_id = db.Column(
        UUIDType(binary=False),
        db.ForeignKey("entity.id"),
        default=None,
        nullable=True,
        index=True,
    )

    __table_args__ = (
        db.UniqueConstraint(
            "name",
            "entity_id",
            "asset_instance_id",
            "output_type_id",
            "task_type_id",
            "temporal_entity_id",
            "representation",
            "revision",
            name="output_file_uc",
        ),
    )

    def __repr__(self):
        return "<OutputFile %s>" % self.id



"""add columns: notifications_mattermost_{enabled,userid}

Revision ID: 3e0538ddf80f
Revises: a66508788c53
Create Date: 2021-11-12 15:31:19.621953

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "3e0538ddf80f"
down_revision = "a66508788c53"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "person",
        sa.Column(
            "notifications_mattermost_enabled", sa.Boolean(), nullable=True
        ),
    )
    op.add_column(
        "person",
        sa.Column(
            "notifications_mattermost_userid",
            sa.String(length=60),
            nullable=True,
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("person", "notifications_mattermost_userid")
    op.drop_column("person", "notifications_mattermost_enabled")
    # ### end Alembic commands ###



"""For SearchFilterGroup.is_shared

Revision ID: 680c64565f9d
Revises: f344b867a911
Create Date: 2024-05-29 02:13:20.764614

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "680c64565f9d"
down_revision = "f344b867a911"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("search_filter_group", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("is_shared", sa.Boolean(), nullable=True)
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("search_filter_group", schema=None) as batch_op:
        batch_op.drop_column("is_shared")

    # ### end Alembic commands ###



from flask import Blueprint

from zou.app.utils.api import configure_api_from_blueprint

from zou.app.blueprints.chats.resources import (
    ChatResource,
    ChatMessagesResource,
    ChatMessageResource,
)


routes = [
    ("/data/entities/<entity_id>/chat", ChatResource),
    ("/data/entities/<entity_id>/chat/messages", ChatMessagesResource),
    (
        "/data/entities/<entity_id>/chat/messages/<chat_message_id>",
        ChatMessageResource,
    ),
]

blueprint = Blueprint("chats", "chats")
api = configure_api_from_blueprint(blueprint, routes)



"""empty message

Revision ID: 3d5c93bafb9d
Revises: 7dc79d4ed7cd
Create Date: 2018-05-30 20:04:38.461178

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "3d5c93bafb9d"
down_revision = "7dc79d4ed7cd"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "project", sa.Column("has_avatar", sa.Boolean(), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("project", "has_avatar")
    # ### end Alembic commands ###



from flask_jwt_extended import jwt_required


from zou.app.models.comment import Comment
from zou.app.models.attachment_file import AttachmentFile

from zou.app.services import (
    comments_service,
    deletion_service,
    notifications_service,
    persons_service,
    tasks_service,
    user_service,
    projects_service,
)
from zou.app.utils import events, permissions

from zou.app.blueprints.crud.base import BaseModelResource, BaseModelsResource


class CommentsResource(BaseModelsResource):
    def __init__(self):
        BaseModelsResource.__init__(self, Comment)


class CommentResource(BaseModelResource):
    def __init__(self):
        BaseModelResource.__init__(self, Comment)
        self.protected_fields += ["mentions", "department_mentions"]

    def get_serialized_instance(self, instance_id, relations=True):
        return tasks_service.get_comment(instance_id, relations=relations)

    def clean_get_result(self, result):
        if permissions.has_client_permissions():
            person = persons_service.get_person(result["person_id"])
            if person["role"] != "client":
                result["text"] = ""
                result["attachment_files"] = []
                result["checklist"] = []
        attachment_files = []
        if (
            "attachment_files" in result
            and len(result["attachment_files"]) > 0
        ):
            for attachment_file_id in result["attachment_files"]:
                attachment_file = AttachmentFile.get(attachment_file_id)
                attachment_files.append(attachment_file.present())
            result["attachment_files"] = attachment_files
        return result

    def pre_update(self, instance_dict, data):
        self.task_status_change = False
        if instance_dict["task_status_id"] != data.get("task_status_id", None):
            self.task_status_change = True
            self.previous_task_status_id = instance_dict["task_status_id"]
        return data

    def post_update(self, instance_dict, data):
        comment = comments_service.reset_mentions(instance_dict)
        if self.task_status_change:
            task_id = comment["object_id"]
            task = tasks_service.reset_task_data(task_id)
            events.emit(
                "task:status-changed",
                {
                    "task_id": task_id,
                    "new_task_status_id": comment["task_status_id"],
                    "previous_task_status_id": self.previous_task_status_id,
                    "person_id": comment["person_id"],
                },
                project_id=task["project_id"],
            )

        tasks_service.clear_comment_cache(comment["id"])
        notifications_service.reset_notifications_for_mentions(comment)
        return comment

    def check_read_permissions(self, instance):
        return user_service.check_comment_access(instance["id"])

    def check_update_permissions(self, instance, data):
        if permissions.has_admin_permissions():
            return True
        else:
            task = tasks_service.get_task(
                instance["object_id"], relations=True
            )
            task_type = tasks_service.get_task_type(task["task_type_id"])
            project = projects_service.get_project(
                task["project_id"], relations=True
            )
            current_user = persons_service.get_current_user(relations=True)
            if current_user["id"] not in project["team"]:
                raise permissions.PermissionDenied

            if permissions.has_manager_permissions():
                return True

            change_pinned = (
                "pinned" in data.keys()
                and data["pinned"] != instance["pinned"]
            )
            change_checklist = (
                "checklist" in data.keys()
                and data["checklist"] != instance["checklist"]
            )
            is_supervisor = permissions.has_supervisor_permissions()
            is_supervisor_in_department = (
                current_user["departments"] == []
                or task_type["department_id"] in current_user["departments"]
            )
            is_assigned = current_user["id"] in task["assignees"]
            comment_from_current_user = (
                current_user["id"] == instance["person_id"]
            )

            if change_pinned and (
                not is_supervisor or not is_supervisor_in_department
            ):
                raise permissions.PermissionDenied

            if change_checklist and (
                not comment_from_current_user
                and (
                    (
                        is_supervisor
                        and not (is_supervisor_in_department or is_assigned)
                    )
                    or (not is_supervisor and not is_assigned)
                )
                and (
                    len(data["checklist"]) == len(instance["checklist"])
                    and all(
                        all(
                            (
                                k == "checked"
                                or data["checklist"][i].get(k) == c[k]
                            )
                            for k in c.keys()
                        )
                        for i, c in enumerate(instance["checklist"])
                    )
                )
            ):
                raise permissions.PermissionDenied

            if (
                not comment_from_current_user
                and len(set(data.keys()) - set(["pinned", "checklist"])) > 0
            ):
                raise permissions.PermissionDenied

            if (
                "person_id" in data.keys()
                and data["person_id"] != current_user["id"]
            ):
                raise permissions.PermissionDenied

            if (
                "object_id" in data.keys()
                and data["object_id"] != instance["object_id"]
            ):
                raise permissions.PermissionDenied

            if "task_status_id" in data.keys():
                user_service.check_task_status_access(data["task_status_id"])

            return True

    def pre_delete(self, comment):
        task = tasks_service.get_task(comment["object_id"])
        self.previous_task_status_id = task["task_status_id"]
        return comment

    def post_delete(self, comment):
        task = tasks_service.get_task(comment["object_id"])
        self.new_task_status_id = task["task_status_id"]
        if self.previous_task_status_id != self.new_task_status_id:
            events.emit(
                "task:status-changed",
                {
                    "task_id": task["id"],
                    "new_task_status_id": self.new_task_status_id,
                    "previous_task_status_id": self.previous_task_status_id,
                    "person_id": comment["person_id"],
                },
                project_id=task["project_id"],
            )
        return comment

    @jwt_required()
    def delete(self, instance_id):
        """
        Delete a comment corresponding at given ID.
        """
        comment = tasks_service.get_comment(instance_id)
        task = tasks_service.get_task(comment["object_id"])
        if permissions.has_manager_permissions():
            user_service.check_project_access(task["project_id"])
        else:
            user_service.check_person_access(comment["person_id"])
        self.pre_delete(comment)
        deletion_service.remove_comment(comment["id"])
        tasks_service.reset_task_data(comment["object_id"])
        tasks_service.clear_comment_cache(comment["id"])
        self.post_delete(comment)
        return "", 204
